import re
import json
from time import time
from data_source.models import DataSource
from cwl_execution.models import Executable, UserJobConfig, ExecutionLog
from cryptography.fernet import Fernet
import cryptography
from django.conf import settings
import os
from utils.command_line_execution import create_command_executor, CommandExecutor
import string
import environ
import paramiko

from heliport.logger import logger
from heliport.utils.environment import get_paramiko_krypto_key
from .cwl_generation import generate_cwl_filename

env = environ.Env()


def save(commandline_argument: str) -> str:
    allowed_symbols = string.digits + string.ascii_letters
    return re.sub(fr'[^{allowed_symbols}]', '_', commandline_argument)


class CWLExecutorManager:
    executor: CommandExecutor
    job_config: UserJobConfig

    def __init__(self, job_config, debug=False, auto_run=True, execution_id=None):
        self.job_config = job_config
        self.execution_id = execution_id
        self.debug = debug
        self.tool_filename = 'tool.cwl'
        self.job_filename = 'job.json'

        self.executor = self.build_ssh_shell()
        self.execution_log = ExecutionLog(config=self.job_config)
        self.return_message = None
        self.pwd = None
        self.start_time = time()
        self.profiling = dict()
        if auto_run:
            self.run()

    def config_invalid(self):
        if self.login is None:
            self.configure_job_error('no ssh connection specified')
            return True

        directory = self.execution_directory
        if directory.startswith('~/'):
            directory = directory[2:]
        invalid_letters = set(directory) - set(string.ascii_letters + string.digits + '_/')
        if invalid_letters:
            error_message = f'invalid letters in directory {repr(self.execution_directory)}: {invalid_letters}'
            self.configure_job_error(error_message)
            return True
        return False

    @staticmethod
    def datasource_has_native_cwl_support(source: DataSource):
        return source.protocol in {'http', 'https'} and source.login is None

    def save_connect(self, executor: CommandExecutor):
        assert not executor.connected, 'executor is already connected'
        try:
            executor.connect()
        except cryptography.fernet.InvalidToken:
            error_message = f'could not decrypt ssh key - try reconnecting the "{self.login.name}" ssh connection'
            self.configure_job_error(error_message)
            return False
        except paramiko.ssh_exception.AuthenticationException:
            error_message = f'could not authenticate - try reconnecting the "{self.login.name}" ssh connection'
            self.configure_job_error(error_message)
            return False
        return True

    @property
    def job(self):
        return self.job_config.job

    @property
    def user(self):
        return self.job_config.ssh_login.user

    @property
    def login(self):
        return self.job_config.ssh_login

    @property
    def execution_directory(self):
        return self.job_config.directory

    @property
    def job_folder(self):
        return os.path.join(self.execution_directory, f'job_{save(self.job.name)}_{self.job.pk}')

    @property
    def absolute_job_folder(self):
        if self.pwd is None:
            raise ValueError('call prepare_job_folder_and_cwltool() first')

        location = self.job_folder
        if location.startswith('~/'):
            location = os.path.join(self.pwd, location[2:])

        return location

    @property
    def srun(self):
        return f'echo "" | srun --time {settings.SRUN_TIMEOUT} ' if settings.USE_SRUN and self.login.slurm else ''

    def configure_job_running(self):
        logger.debug('set job status to running')
        self.job.output_text = f'execution by {self.user.display_name}\n...\n'
        self.job.output_code = self.job.OutputCodes.WARNING
        self.job.status = self.job.StatusCodes.RUNNING
        self.job.save()

    def configure_job_error(self, message='an undefined error occurred'):
        self.job.output_text = message
        self.job.output_code = self.job.OutputCodes.ERROR
        self.job.status = self.job.StatusCodes.STOPPED
        self.job.save()

    def configure_job_stopped(self, status_code=0):
        self.job.output_code = self.job.OutputCodes.ERROR
        self.job.status = self.job.StatusCodes.STOPPED

        if status_code == 0:
            self.job.output_code = self.job.OutputCodes.MESSAGE
        self.job.save()

    def append_job_output_text(self, message):
        self.job.output_text += message
        self.job.save()

    def execute(self, command: str):
        return self.executor.execute(command)

    def transfer_cwl(self, cwl, relative_location):
        file_path = os.path.join(self.absolute_job_folder, relative_location)
        logger.debug(file_path)
        with self.executor.file(file_path, 'w') as f:
            print(json.dumps(cwl), file=f)

    def run(self):
        if self.config_invalid():
            self.return_message = 'config invalid'
            return self.return_message
        self.profiling['validate config'] = time()

        logger.debug('STARTING JOB EXECUTION')
        self.configure_job_running()
        self.execution_log.start(self.execution_id)
        if not self.save_connect(self.executor):
            self.return_message = 'ssh connection failed'
            return self.return_message
        self.profiling['connect to cluster'] = time()

        with self.executor:
            logger.debug('calling setup command')
            self.execute(self.job.setup_command)

            logger.debug('preparing environment on cluster')
            self.prepare_job_folder_and_cwltool()
            self.profiling['setup environment'] = time()

            logger.debug('transferring files')
            tool_cwl = self.extract_pure_executable_cwl(self.job.base_executable)
            self.transfer_cwl(tool_cwl, self.tool_filename)
            for exe in self.job.project.executable_set.all():
                name = generate_cwl_filename(exe.cwl_json, exe)
                cwl = self.extract_pure_executable_cwl(exe)
                self.transfer_cwl(cwl, name)

            job_cwl, data_sources = self.extract_pure_job_cwl_and_data_sources()
            self.transfer_cwl(job_cwl, self.job_filename)
            for data_source in data_sources:
                self.transfer_data_source(data_source)
            self.profiling['transfer files'] = time()

            logger.debug('run cwltool')
            status = self.run_cwl_tool()
            self.profiling['execute workflow'] = time()

        self.configure_job_stopped(status)
        self.save_metadata()
        self.return_message = 'success' if status == 0 else 'cwl error'
        self.profiling['collect and store metadata'] = time()
        self.save_profiling()
        return self.return_message

    def save_profiling(self):
        last_time = self.start_time
        deltas = dict()
        for step, seconds in self.profiling.items():
            deltas[step] = seconds - last_time
            last_time = seconds

        self.execution_log.step_times = json.dumps(deltas)
        self.execution_log.save()

    def save_metadata(self):
        find_memory_usage = re.compile(r'\[([^]\[]+)] Max memory used: ([^ ]+)')
        memory_usage = {}
        for line in self.job.output_text.splitlines():
            match = find_memory_usage.search(line)
            if match:
                memory_usage[match[1]] = match[2]

        self.execution_log.memory_usage = json.dumps(memory_usage)
        self.execution_log.stop()

    def run_cwl_tool(self):
        debug_cmd = '--debug ' if self.debug else ''
        cmd = f'{self.srun}cwltool {debug_cmd}{self.tool_filename} {self.job_filename}'
        status = self.execute(cmd).status
        return status

    def transfer_data_source(self, data_source):
        if data_source.uri.startswith('ssh://'):
            self.transfer_ssh_data_source(data_source)
        elif not self.datasource_has_native_cwl_support(data_source):
            self.download_data_with_curl(data_source)

    def download_data_with_curl(self, data_source: DataSource):
        if data_source.login is None:
            login_info_command = ''

        elif data_source.login.type == data_source.login.LoginTypes.USER_PASSWORD:
            login_info_command = f'--user "$USERNAME:$PASSWORD" '
            uname = data_source.login.username
            password = data_source.login.password
            self.execute(f'read -N {len(uname)} -r USERNAME\n{uname}')
            self.execute(f'read -N {len(password)} -r -s PASSWORD\n{password}')

        elif data_source.login.type == data_source.login.LoginTypes.TOKEN:
            login_info_command = f'-H "$TOKEN_HEADER" '
            token = data_source.login.key
            self.execute(f'read -N {len(token)} -r -s TOKEN_HEADER\n{token}')

        else:
            login_info_command = f'--unsupported_login_type "{data_source.login.type}"'

        data_path = self.data_source_filename(data_source)
        logger.debug(data_path)
        self.execute(f'curl {login_info_command}"{data_source.uri}" > "{data_path}"')

    def transfer_ssh_data_source(self, data_source: DataSource):
        if data_source.login is None:
            self.execute(f'echo "ERROR: no ssh login for {save(data_source.name)}"')
            return

        self.execute(f'echo "Transferring ssh data source: {save(data_source.name)}"')
        remote_login = data_source.login.build_command_executor()
        if not self.save_connect(remote_login):
            self.execute(f'echo "ERROR: failed to connect to {save(data_source.name)}"')

        file_name = os.path.join(self.absolute_job_folder, self.data_source_filename(data_source))
        logger.debug(file_name)
        with remote_login, self.executor.file(file_name, 'w') as destination:
            source = remote_login.file(data_source.path)
            data = source.read(settings.SSH_DATA_TRANSFER_BATCH_SIZE)
            while data:
                destination.write(data)
                data = source.read(settings.SSH_DATA_TRANSFER_BATCH_SIZE)

    def extract_pure_job_cwl_and_data_sources(self):
        job_cwl = self.job.cwl
        data_sources = []
        for param_id, param in job_cwl.items():
            data_source = self.extract_data_source_from_cwl_parameter(param)
            if data_source is None:
                continue

            data_sources.append(data_source)
            uri = self.data_source_filename(data_source)
            param['location'] = uri

        return job_cwl, data_sources

    @classmethod
    def data_source_filename(cls, data_source: DataSource):
        uri = data_source.uri
        if not cls.datasource_has_native_cwl_support(data_source):
            uri = save(uri)
        return uri

    @staticmethod
    def extract_data_source_from_cwl_parameter(param):
        if not isinstance(param, dict):
            return None
        if param.get('class') != 'File':
            return None
        data_source_id = param.get('location', '').split('_')[0]
        try:
            data_source_id_int = int(data_source_id)
        except ValueError:
            return None
        try:
            data_source = DataSource.objects.get(pk=data_source_id_int)
        except DataSource.DoesNotExist:
            return None

        return data_source

    def extract_pure_executable_cwl(self, exe):
        tool_cwl = exe.cwl
        if self.job.base_executable.type == Executable.ExecutableTypes.WORKFLOW:
            self.update_renamed_workflow_steps(tool_cwl)
        return tool_cwl

    @staticmethod
    def update_renamed_workflow_steps(tool_cwl):
        regex = re.compile(r'.*_(\d+).cwl$')
        for step in tool_cwl.get('steps', []):
            run = step.get('run', '')
            match = regex.match(run)
            try:
                exe_id = int(match.group(1))
            except AttributeError:
                continue
            try:
                exe = Executable.objects.get(pk=exe_id)
            except Executable.DoesNotExist:
                continue
            step['run'] = generate_cwl_filename(exe.cwl, exe)

    def prepare_job_folder_and_cwltool(self):
        test_command = f'source {self.execution_directory}/venv/bin/activate && pip list --format=legacy | grep cwltool'
        cwl_installed = self.execute(test_command).status == 0

        if not cwl_installed:
            self.execute('echo "CWL TOOL NOT IN VENV - GOING TO CREATE VENV AND INSTALL CWL TOOL"')
            self.execute(f'python -m venv {self.execution_directory}/venv')
            self.execute(f'source {self.execution_directory}/venv/bin/activate')
            self.execute(f'pip install cwltool')

        self.execute(f'mkdir -p {self.job_folder}')
        self.pwd = self.execute('pwd').last_complete_line
        self.execute(f'cd {self.job_folder}')

    def build_ssh_shell(self):
        if self.login.type != self.login.LoginTypes.SSH:
            raise ValueError("not a ssh login")

        connection = create_command_executor(
            host=self.login.machine,
            username=self.login.username,
            key=self.login.key,
            key_password_callback=get_paramiko_krypto_key,
            indirection=self.login.build_command_executor_indirection(),
            shell=True,
            auto_connect=False,
            default_result_callback=self.append_job_output_text
        )

        return connection
