from typing import List, Dict, Union, Any
from collections import namedtuple
import json
import yaml

from .cwl_parameter import CwlParameter, ParameterValue


class CwlJob:
    def __init__(self, parameters: List[CwlParameter],
                 job_dict: Dict[str, Union[Any, ParameterValue]] = None,
                 prefer_json=False):
        if job_dict is None:
            job_dict = dict()
        self.prefer_json = prefer_json
        self.original_parameters = parameters

        self.dict = dict()
        self.unsupported_dict = dict()

        for parameter in parameters:
            p = job_dict.get(parameter.name)
            if parameter.supported:
                self.dict[parameter.name] = parameter.generate_parameter_value(p)
            elif isinstance(p, ParameterValue):
                self.unsupported_dict[parameter.name] = p.cwl
            else:
                self.unsupported_dict[parameter.name] = p

    def get_value(self, name):
        parameter_value = self.dict.get(name)
        if parameter_value is None:
            return None
        else:
            return parameter_value.value

    @property
    def is_valid(self):
        return all(p.is_valid for p in self.dict.values())

    @property
    def cwl(self):
        result = {key: value.cwl for key, value in self.dict.items()}
        result.update(self.unsupported_dict)
        return result

    @property
    def json(self):
        return json.dumps(self.cwl, indent=3)

    @property
    def yaml(self):
        return yaml.safe_dump(self.cwl)

    @property
    def cwl_str(self):
        if self.prefer_json:
            return self.json
        else:
            return self.yaml

    @property
    def unsupported_cwl_str(self):
        if self.prefer_json:
            return json.dumps(self.unsupported_dict)
        else:
            return yaml.safe_dump(self.unsupported_dict)

    @property
    def has_parameters(self):
        return bool(self.dict or self.unsupported_dict)

    @property
    def parameter_details(self):
        details = namedtuple('ParameterDetails', ['name', 'description', 'type', 'value', 'supported_type'])
        return [
            details(
                name=p.name,
                description=p.description,
                type=p.type,
                supported_type=p.supported,
                value=self.get_value(p.name)
            )
            for p in self.original_parameters
        ]
