import abc
import typing

from data_source.models import DataSource


class CwlParameter:
    def __init__(self, name, value_type, description=''):
        self.name = name
        self.type = value_type
        self.description = description

    def generate_parameter_value(self, value=None):
        return supported_parameter_types[self.type](value)

    @property
    def supported(self):
        return self.type in supported_parameter_types


class ParameterValue:
    @property
    @abc.abstractmethod
    def cwl(self):
        pass

    @property
    @abc.abstractmethod
    def is_valid(self):
        pass

    @property
    @abc.abstractmethod
    def value(self):
        pass


class StringParameterValue(ParameterValue):
    def __init__(self, value=None):
        if value is None:
            self._value = None
        else:
            self._value = str(value)

    @property
    def cwl(self):
        assert self.is_valid
        return self._value

    @property
    def is_valid(self):
        return isinstance(self._value, str)

    @property
    def value(self):
        return self._value


class BooleanParameterValue(ParameterValue):
    def __init__(self, value=None):
        if value is None:
            self._value = None
        elif isinstance(value, str):
            self._value = value.lower() == 'true'
        else:
            self._value = bool(value)

    @property
    def cwl(self):
        assert self.is_valid
        return self._value

    @property
    def is_valid(self):
        return isinstance(self._value, bool)

    @property
    def value(self):
        return self._value


class FileParameterValue(ParameterValue):
    data_source: typing.Optional[DataSource]

    def __init__(self, data_source=None):
        if data_source is None:
            self.data_source = None

        elif isinstance(data_source, DataSource):
            self.data_source = data_source

        elif isinstance(data_source, dict):
            assert data_source.get('class') == 'File'
            assert 'location' in data_source
            assert '_' in data_source['location'], 'location has the following structure: <pk>_<uri>'
            pk = data_source['location'].split('_')[0]
            self.data_source = DataSource.objects.get(pk=pk)

        else:
            self.data_source = DataSource.objects.get(pk=data_source)

    @property
    def cwl(self):
        assert self.is_valid
        return {
            'class': 'File',
            'location': f'{self.data_source.pk}_{self.data_source.uri}'
        }

    @property
    def is_valid(self):
        return isinstance(self.data_source, DataSource)

    @property
    def value(self):
        if self.data_source is None:
            return None
        else:
            return self.data_source.pk


supported_parameter_types = {
    'string': StringParameterValue,
    'boolean': BooleanParameterValue,
    'File': FileParameterValue
}
