import json

from django.db import models
from django.utils import timezone

from heliport.models import Project, LoginInfo
from .cwl_utils.cwl_parameter import CwlParameter


def json_load_dict(json_string, error_dict=None):
    if error_dict is None:
        error_dict = dict()

    try:
        cwl = json.loads(json_string)
    except ValueError:
        cwl = error_dict
    except TypeError:
        cwl = error_dict

    if not isinstance(cwl, dict):
        cwl = error_dict

    return cwl


class Executable(models.Model):
    class ExecutableTypes(models.IntegerChoices):
        TOOL = 1
        WORKFLOW = 2
        ARBITRARY_CWL = 3

    class CWLTypes(models.IntegerChoices):
        YAML = 2
        JSON = 1

    executable_id = models.AutoField(primary_key=True)
    cwl_json = models.TextField(blank=True, default="")
    cwl_type = models.IntegerField(choices=CWLTypes.choices, default=CWLTypes.JSON)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    type = models.IntegerField(choices=ExecutableTypes.choices)

    def uses_files(self):
        try:
            cwl = json_load_dict(self.cwl_json, {'inputs': {}})
            inputs = cwl['inputs']
            if isinstance(inputs, dict):
                inputs = inputs.values()
            return any(param['type'] == 'File' for param in inputs)
        except KeyError:
            return False

    @property
    def cwl(self):
        return json_load_dict(self.cwl_json, {'label': 'SYNTAX ERROR', 'doc': 'SYNTAX ERROR IN JSON'})

    @property
    def job_count(self):
        return len(self.job_set.all())

    @property
    def parameters(self):
        parameters = self.cwl.get('inputs', [])
        if isinstance(parameters, dict):
            parameters = [{'id': key, **values} for key, values in parameters.items()]

        return [
            CwlParameter(p.get('id', 'INVALID'), p.get('type', 'no_type'), p.get('label', ''))
            for p in parameters
        ]


class Job(models.Model):
    class OutputCodes(models.IntegerChoices):
        UNDEFINED = -1
        MESSAGE = 0
        WARNING = 1
        ERROR = 2

    class StatusCodes(models.TextChoices):
        RUNNING = "Running"
        PAUSED = "Paused"
        STOPPED = "Stopped"

    job_id = models.AutoField(primary_key=True)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    name = models.CharField(max_length=100)
    status = models.CharField(max_length=80, choices=StatusCodes.choices, default=StatusCodes.STOPPED)
    description = models.TextField()
    output_text = models.TextField()
    output_code = models.IntegerField(choices=OutputCodes.choices, default=OutputCodes.UNDEFINED)
    cwl_json = models.TextField(blank=True, default="")
    repeat = models.BooleanField(default=False)
    interval = models.IntegerField(default=0)
    base_executable = models.ForeignKey(Executable, on_delete=models.CASCADE)
    setup_command = models.TextField(blank=True)

    def __str__(self):
        return str(self.name)

    @property
    def cwl(self):
        return json_load_dict(self.cwl_json, {"SYNTAX ERROR": "SYNTAX ERROR IN JSON"})


class UserJobConfig(models.Model):
    user_job_config_id = models.AutoField(primary_key=True)
    job = models.ForeignKey(Job, on_delete=models.CASCADE)
    ssh_login = models.ForeignKey(LoginInfo, on_delete=models.SET_NULL, null=True)
    directory = models.CharField(max_length=300)

    def __str__(self):
        return f'{self.job} config ({self.pk})'

    @property
    def project(self):
        return self.job.project


class ExecutionLog(models.Model):
    config = models.ForeignKey(UserJobConfig, on_delete=models.CASCADE)
    start_time = models.DateTimeField()
    end_time = models.DateTimeField(null=True, blank=True)
    memory_usage = models.TextField(default='')
    execution_id = models.TextField(null=True, blank=True)
    aborted = models.BooleanField(default=False)
    step_times = models.TextField(null=True, blank=True)

    def __str__(self):
        return f'{str(self.config.job)} log ({self.pk})'

    def start(self, execution_id=None):
        self.start_time = timezone.now()
        self.execution_id = execution_id
        self.save()

    def stop(self, kill=False):
        self.end_time = timezone.now()
        self.aborted = kill
        self.save()

    @property
    def duration(self):
        if self.end_time is not None:
            return self.end_time - self.start_time
        return timezone.now() - self.start_time

    @property
    def seconds_since_start(self):
        return (timezone.now() - self.start_time).total_seconds()

    @property
    def duration_str(self):
        total_seconds = int(self.duration.total_seconds())

        days = total_seconds // 86400
        remaining_hours = total_seconds % 86400
        remaining_minutes = remaining_hours % 3600
        hours = remaining_hours // 3600
        minutes = remaining_minutes // 60
        seconds = remaining_minutes % 60

        days_str = f'{days}d ' if days else ''
        hours_str = f'{hours}h ' if hours else ''
        minutes_str = f'{minutes}m ' if minutes else ''
        seconds_str = f'{seconds}s' if seconds and not hours_str else ''

        return f'{days_str}{hours_str}{minutes_str}{seconds_str}'

    @property
    def memory_usage_dict(self):
        return json_load_dict(self.memory_usage)

    @property
    def profiling(self):
        if self.step_times:
            return json.loads(self.step_times)
        return None
