import yaml
from collections import defaultdict
from django.db.models import Q
from ansi2html.converter import Ansi2HTMLConverter
from django.shortcuts import redirect, get_object_or_404, render, HttpResponse
from django.http import JsonResponse, HttpResponseForbidden
from django.utils import translation
from django.conf import settings
from rest_framework import viewsets
from rest_framework.views import Response as RestResponse
from rest_framework import permissions as rest_permissions
from rest_framework.decorators import action
from heliport_project.celery import app as celery_app

from .serializers import ExecutableSerializer, JobSerializer, UserJobConfigSerializer
from heliport.permissions import HeliportObjectPermission, is_object_member
from .models import *
from heliport.custom_views import HeliportProjectView, HeliportObjectView, HeliportLoginRequiredMixin, django_views
from .cwl_utils import cwl_generation, cwl_execution, cwl_tool, cwl_workflow
from .cwl_utils.cwl_job import CwlJob
from heliport.logger import logger
from . import tasks


class ArbitraryCWLListView(HeliportProjectView, django_views.ListView):
    model = Executable
    template_name = 'cwl_execution/arbitrary_cwl.html'
    context_object_name = "arbitrary_cwls"

    def get_queryset(self):
        project = get_object_or_404(Project, pk=self.kwargs['project'])
        return Executable.objects.filter(type=Executable.ExecutableTypes.ARBITRARY_CWL, project=project)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["cwl_types"] = Executable.CWLTypes.choices
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        cwl_type = request.POST.get('cwl_type')
        cwl = request.POST.get('cwl')
        executable_id = request.POST.get('remove')

        if executable_id:
            executable = get_object_or_404(Executable, pk=executable_id)
            executable.delete()
        elif cwl:
            cwl_type = int(cwl_type)
            super().get(request, *args, **kwargs)
            context = self.get_context_data()
            context['cwl_type'] = cwl_type
            context['cwl'] = cwl

            if cwl_type == Executable.CWLTypes.JSON:
                try:
                    cwl = json.loads(cwl)
                    if not isinstance(cwl, dict):
                        context['message'] = f'expected dict as cwl but got {type(cwl)}'
                        return render(request, self.template_name, context)
                except ValueError as e:
                    context['message'] = f'JSON parse failed: {e}'
                    return render(request, self.template_name, context)

            elif cwl_type == Executable.CWLTypes.YAML:
                try:
                    cwl = yaml.safe_load(cwl)
                    if not isinstance(cwl, dict):
                        context['message'] = f'expected dict as cwl but got {type(cwl)}'
                        return render(request, self.template_name, context)
                except yaml.YAMLError as e:
                    context['message'] = f'YAML parse failed: {e}'
                    return render(request, self.template_name, context)
            else:
                return render(request, self.template_name, context)

            Executable(
                project=context['project'],
                type=Executable.ExecutableTypes.ARBITRARY_CWL,
                cwl_type=cwl_type,
                cwl_json=json.dumps(cwl)
            ).save()

        return redirect('cwl_execution:arbitrary_cwl_list', project=self.kwargs['project'])


class ArbitraryCWLUpdateView(HeliportObjectView, django_views.UpdateView):
    model = Executable
    template_name = "cwl_execution/arbitrary_cwl.html"
    fields = ["cwl_json", "cwl_type"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = context['project']
        context["arbitrary_cwls"] = project.executable_set.filter(type=Executable.ExecutableTypes.ARBITRARY_CWL)
        context["cwl_types"] = Executable.CWLTypes.choices
        context["cwl_type"] = self.object.cwl_type

        cwl_dict = json_load_dict(self.object.cwl_json, {
            "baseCommand": ["echo", "INVALID SYNTAX! This Arbitrary CWL does not work because of a syntax error."],
            "class": "CommandLineTool",
            "cwlVersion": "v1.0",
            "inputs": [],
            "outputs": [],
            "label": "SYNTAX ERROR",
            "doc": "SYNTAX ERROR",
        })

        if self.object.cwl_type == Executable.CWLTypes.YAML:
            context["cwl"] = yaml.dump(cwl_dict)
        else:
            context["cwl"] = json.dumps(cwl_dict, indent=3)

        context["update"] = True
        return context

    def post(self, request, *args, **kwargs):
        cwl_type = request.POST.get("cwl_type")
        cwl = request.POST.get("cwl")

        if cwl:
            cwl_type = int(cwl_type)
            super().get(request, *args, **kwargs)
            context = self.get_context_data()
            context["cwl_type"] = cwl_type
            context["cwl"] = cwl

            if cwl_type == Executable.CWLTypes.JSON:
                try:
                    cwl = json.loads(cwl)
                    if not isinstance(cwl, dict):
                        context["message"] = f"expected dict as cwl but got {type(cwl)}"
                        return render(request, self.template_name, context)
                except ValueError as e:
                    context["message"] = f"JSON parse failed: {e}"
                    return render(request, self.template_name, context)

            elif cwl_type == Executable.CWLTypes.YAML:
                try:
                    cwl = yaml.safe_load(cwl)
                    if not isinstance(cwl, dict):
                        context["message"] = f"expected dict as cwl but got {type(cwl)}"
                        return render(request, self.template_name, context)
                except yaml.YAMLError as e:
                    context["message"] = f"YAML parse failed: {e}"
                    return render(request, self.template_name, context)
            else:
                return render(request, self.template_name, context)

            executable = self.get_object()
            executable.cwl_json = json.dumps(cwl)
            executable.cwl_type = cwl_type
            executable.save()

        return redirect('cwl_execution:arbitrary_cwl_list', project=self.get_object().project.pk)


class ToolListView(HeliportProjectView, django_views.ListView):
    model = Executable
    template_name = 'cwl_execution/tool_list.html'
    context_object_name = 'tools'

    def get_queryset(self):
        project = get_object_or_404(Project, pk=self.kwargs['project'])
        return project.executable_set.filter(type=Executable.ExecutableTypes.TOOL)

    def post(self, request, *args, **kwargs):
        tool_to_delete = get_object_or_404(Executable, pk=request.POST.get('delete'))
        project = tool_to_delete.project
        tool_to_delete.delete()
        return redirect('cwl_execution:tool_list', project=project.pk)


class ExecutableRawYamlView(HeliportObjectView, django_views.DetailView):
    model = Executable

    def get(self, request, *args, **kwargs):
        exe = self.get_object()

        try:
            generated_cwl_yml = yaml.safe_dump(json.loads(exe.cwl_json))
        except ValueError as e:
            generated_cwl_yml = f"cannot convert json to yaml: {e}\n\n{exe.cwl}"

        return HttpResponse(generated_cwl_yml, content_type="text/plain")


class ToolJsonDescriptionView(HeliportObjectView, django_views.DetailView):
    model = Executable

    def get(self, request, *args, **kwargs):
        obj = self.get_object()
        description = cwl_tool.cwl_to_tool_description(obj.cwl)
        return JsonResponse(description)


class ToolUpdateView(HeliportObjectView, django_views.DetailView):
    template_name = 'cwl_execution/tool_edit.html'
    model = Executable

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['edit_mode'] = True
        return context

    def post(self, request, *args, **kwargs):
        tool = self.get_object()
        tool_description = json.loads(request.POST.get('tool_json', '{}'))
        tool_cwl = cwl_tool.tool_description_to_cwl(tool_description)
        tool.cwl_json = json.dumps(tool_cwl)
        tool.save()

        return redirect('cwl_execution:tool_list', project=tool.project.pk)


class ToolCreateView(HeliportProjectView, django_views.TemplateView):
    template_name = "cwl_execution/tool_edit.html"

    def post(self, request, *args, **kwargs):
        project = get_object_or_404(Project, pk=kwargs['project'])
        new_tool = Executable(project=project, type=Executable.ExecutableTypes.TOOL)
        tool_description = json.loads(request.POST.get('tool_json', '{}'))
        tool_cwl = cwl_tool.tool_description_to_cwl(tool_description)
        new_tool.cwl_json = json.dumps(tool_cwl)
        new_tool.save()

        return redirect('cwl_execution:tool_list', project=project.pk)


class WorkflowListView(HeliportProjectView, django_views.ListView):
    model = Executable
    template_name = "cwl_execution/workflow_list.html"
    context_object_name = 'workflows'

    def get_queryset(self):
        project = get_object_or_404(Project, pk=self.kwargs['project'])
        return project.executable_set.filter(type=Executable.ExecutableTypes.WORKFLOW)

    def post(self, request, *args, **kwargs):
        workflow = request.POST.get("delete")
        project = get_object_or_404(Project, pk=self.kwargs['project'])
        if workflow:
            exe = get_object_or_404(Executable, pk=workflow)
            exe.delete()

        return redirect('cwl_execution:workflow_list', project=project.pk)


class VisualizationDataView(HeliportObjectView, django_views.DetailView):
    model = Executable

    def get(self, request, *args, **kwargs):
        exe = self.get_object()
        full_workflow = request.GET.get('full_workflow')
        if full_workflow is not None and full_workflow.lower() != 'false':
            if exe.type != Executable.ExecutableTypes.WORKFLOW:
                logger.error(f'accessed "full_workflow" visualization data vor type={exe.type}')
            data = cwl_workflow.cwl_to_visualization_data(exe.cwl)
        else:
            data = cwl_workflow.data_for_visualization_node(exe)
        return JsonResponse(data)


class WorkflowUpdateView(HeliportObjectView, django_views.DetailView):
    model = Executable
    context_object_name = "workflow"
    template_name = "cwl_execution/workflow_edit.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['edit_mode'] = True
        query = self.request.GET.get("q")

        executables = set(Executable.objects.filter(project=self.object.project))
        if query:
            context["query"] = query
            executables = cwl_generation.filter_executables(executables, query)

        context["executables"] = list(executables)

        return context

    def post(self, request, *args, **kwargs):
        workflow = self.get_object()
        visualization_data = json.loads(request.POST.get("graph_cwl", "{}"))
        name = request.POST.get("name")
        description = request.POST.get("description")

        cwl = cwl_workflow.visualization_data_to_cwl(visualization_data, name, description)
        workflow.cwl_json = json.dumps(cwl)
        workflow.save()
        return redirect('cwl_execution:workflow_list', project=workflow.project.pk)


class WorkflowCreateView(HeliportProjectView, django_views.TemplateView):
    model = Executable
    context_object_name = "workflow"
    template_name = "cwl_execution/workflow_edit.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        query = self.request.GET.get("q")

        project = get_object_or_404(Project, pk=self.kwargs['project'])
        executables = set(Executable.objects.filter(project=project))
        if query:
            context["query"] = query
            executables = cwl_generation.filter_executables(executables, query)
        context["executables"] = list(executables)

        return context

    def post(self, request, *args, **kwargs):
        name = request.POST.get('name', '')
        description = request.POST.get('description')
        visualization_data = json.loads(request.POST.get('graph_cwl', '{}'))
        project = get_object_or_404(Project, pk=self.kwargs['project'])

        cwl = cwl_workflow.visualization_data_to_cwl(visualization_data, name, description)

        # create new workflow
        Executable(
            project=project,
            type=Executable.ExecutableTypes.WORKFLOW,
            cwl_json=json.dumps(cwl)
        ).save()

        return redirect('cwl_execution:workflow_list', project=project.pk)


class JobCreateView(HeliportObjectView, django_views.DetailView):
    template_name = "cwl_execution/job_create.html"
    model = Executable
    pk_url_kwarg = 'executable'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if 'pk' in self.kwargs:
            job = Job.objects.get(pk=self.kwargs['pk'])
            assert job.base_executable == self.object
            context['job'] = job
            context['update'] = True

            context['job_setup_command'] = job.setup_command
            default_values = job.cwl
        else:
            context['job_setup_command'] = settings.DEFAULT_JOB_SETUP_COMMAND
            default_values = None

        context['cwl_job'] = CwlJob(self.object.parameters, default_values)

        user = self.request.user.heliportuser
        context['ssh_logins'] = user.logininfo_set.filter(type=LoginInfo.LoginTypes.SSH)

        return context

    def post(self, request, *args, **kwargs):
        executable = self.get_object()
        if 'pk' in self.kwargs:
            job_object = Job.objects.get(pk=self.kwargs['pk'])
        else:
            job_object = Job(
                project=executable.project,
                output_text="Successfully created a scientific workflow. To start it press the play button",
                output_code=Job.OutputCodes.MESSAGE,
                base_executable=executable,
            )
        user_inputs = cwl_generation.separate_prefix(request.POST)
        cwl_generation.set_attrs_job(job_object, user_inputs['job'])
        cwl_job = CwlJob(
            executable.parameters,
            user_inputs['value'],
            prefer_json=request.POST.get('cwl_type') == executable.CWLTypes.JSON
        )
        generated_cwl = cwl_job.cwl

        if 'cwl' in request.POST:
            arbitrary_cwl, message = cwl_generation.parse_user_cwl(request.POST['cwl'], request.POST.get('cwl_type'))
            if message is not None:
                super().get(request, *args, **kwargs)
                context = self.get_context_data()
                context['message'] = message
                context['job'] = job_object
                context['cwl_job'] = cwl_job
                context['cwl'] = request.POST.get('cwl')
                context['job_setup_command'] = job_object.setup_command
                return render(request, self.template_name, context)
            generated_cwl.update(arbitrary_cwl)

        job_object.cwl_json = json.dumps(generated_cwl)
        job_object.save()

        return redirect('cwl_execution:job_list', project=executable.project.pk)


class JobListView(HeliportProjectView, django_views.ListView):
    template_name = "cwl_execution/job_list.html"
    model = UserJobConfig
    context_object_name = 'userjobconfigs'

    def get_queryset(self):
        project = get_object_or_404(Project, pk=self.kwargs['project'])
        user = self.request.user.heliportuser
        return UserJobConfig.objects.filter(
            Q(job__project=project) & (Q(ssh_login__user=user) | Q(ssh_login__isnull=True)))

    def get_context_data(self, **kwargs):
        project = Project.objects.get(pk=self.kwargs['project'])
        user = self.request.user.heliportuser

        default_login = None
        logins = LoginInfo.objects.filter(user=user, type=LoginInfo.LoginTypes.SSH)
        if len(logins) == 1:
            default_login = logins.first()
        else:
            last_logins = ExecutionLog.objects.filter(
                start_time__gt=timezone.now() - timezone.timedelta(days=20),
                end_time__isnull=False,
                aborted=False,
                config__ssh_login__user=user
            ).values_list('config__ssh_login', flat=True)
            histogram = defaultdict(int)
            for login in last_logins:
                histogram[login] += 1
            threshold = len(last_logins) * .9
            for login, count in histogram.items():
                if count > threshold:
                    default_login = LoginInfo.objects.get(pk=login)

        for job in project.job_set.all():
            if UserJobConfig.objects.filter(
                    Q(job=job) & (Q(ssh_login__user=user) | Q(ssh_login__isnull=True))).first() is None:
                UserJobConfig(job=job, directory="~/heliport_jobs", ssh_login=default_login).save()

        context = super().get_context_data(**kwargs)
        running = [job.pk for job in project.job_set.all()
                   if job.status == job.StatusCodes.RUNNING]
        context["running"] = running

        user = self.request.user.heliportuser
        context['ssh_logins'] = user.logininfo_set.filter(type=LoginInfo.LoginTypes.SSH)
        return context

    def post(self, request, *args, **kwargs):
        job = get_object_or_404(Job, pk=request.POST['delete'])
        job.delete()

        return redirect('cwl_execution:job_list', project=self.kwargs['project'])


class JobRawYamlView(HeliportObjectView, django_views.DetailView):
    model = Job

    def get(self, request, *args, **kwargs):
        job = self.get_object()

        try:
            raw_yml = yaml.safe_dump(json.loads(job.cwl_json))
        except ValueError as e:
            raw_yml = f"cannot convert json to yaml: {e}\n\n{job.cwl_json}"

        return HttpResponse(raw_yml, content_type="text/plain")


class JobDetailView(HeliportObjectView, django_views.DetailView):
    model = Job
    template_name = 'cwl_execution/job_detail.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        converter = Ansi2HTMLConverter(dark_bg=False)
        context['output_text'] = converter.convert(self.object.output_text, full=False)
        context['execution_count'] = ExecutionLog.objects.filter(config__job=self.object).count()
        context['running_executions'] = ExecutionLog.objects.filter(config__job=self.object, end_time__isnull=True)
        context['last_execution'] = (ExecutionLog.objects
                                     .filter(config__job=self.object, end_time__isnull=False)
                                     .order_by('-start_time').first())

        names = (ExecutionLog.objects
                 .filter(config__job=self.object)
                 .values_list('config__ssh_login__user__display_name', flat=True))
        name_counts = defaultdict(int)
        for name in names:
            name_counts[name] += 1
        context['collaborators'] = sorted([(count, name) for name, count in name_counts.items()], reverse=True)
        return context


class JobOutputView(HeliportObjectView, django_views.DetailView):
    model = Job

    def get(self, request, *args, **kwargs):
        job = self.get_object()
        converter = Ansi2HTMLConverter(dark_bg=False)
        html = converter.convert(job.output_text, full=False)
        return HttpResponse(html)


def start_job(job_config, request, debug=False):
    login = request.GET.get('login')
    directory = request.GET.get('directory')
    if login is not None and login != 'null':
        job_config.ssh_login = LoginInfo.objects.get(pk=login)
    if directory is not None:
        job_config.directory = directory
    job_config.save()

    job = job_config.job
    user = request.user.heliportuser
    if job_config.ssh_login is None:
        job.output_code = job_config.job.OutputCodes.ERROR
        job.output_text = f'Permission denied: no login selected (access by {user})'
        job.save()
        return HttpResponseForbidden()

    if job_config.ssh_login.user != user:
        job.output_code = job_config.job.OutputCodes.ERROR
        job.output_text += f'\nPermission denied: {user} used login by {job_config.ssh_login.user}'
        job.save()
        return HttpResponseForbidden()

    if debug:
        logger.info(f'DEBUGGING JOB {job_config.job.pk}')
    else:
        logger.info(f'STARTING JOB {job.pk}')
    if settings.USE_CELERY:
        tasks.job_start.delay(job_config.pk, debug=debug)
    else:
        status = cwl_execution.CWLExecutorManager(job_config, debug=debug, auto_run=True).return_message
        if debug:
            logger.debug(f'finished debugging {job_config.job} with id={job_config.job.pk} status={status}')
        else:
            logger.debug(f'finished {job} with {status}')

    return redirect('cwl_execution:job_list', project=job.project.pk)


class JobStartView(HeliportObjectView, django_views.DetailView):
    model = UserJobConfig

    def get(self, request, *args, **kwargs):
        return start_job(self.get_object(), self.request)


class JobDebugView(HeliportObjectView, django_views.DetailView):
    model = UserJobConfig

    def get(self, request, *args, **kwargs):
        return start_job(self.get_object(), self.request, debug=True)


class JobStopView(HeliportObjectView, django_views.DetailView):
    model = UserJobConfig

    def get(self, request, *args, **kwargs):
        job = self.get_object().job
        job.status = job.StatusCodes.STOPPED
        running = ExecutionLog.objects.filter(end_time__isnull=True, config__job=job)

        for log in running:
            logger.debug(f'{request.user.heliportuser.display_name} killed {job.name} job')
            if log.execution_id:
                celery_app.control.revoke(log.execution_id, terminate=True)
            log.stop(kill=True)
            job.output_text += f'\n{translation.gettext("killed by")} {request.user.heliportuser.display_name}.\n'
        job.save()
        return redirect('cwl_execution:job_list', project=job.project.pk)


class RunningJobsView(HeliportProjectView, django_views.DetailView):
    model = Project
    pk_url_kwarg = 'project'

    def get(self, request, *args, **kwargs):
        project = self.get_object()

        return JsonResponse({"running": [job.pk for job in project.job_set.all()
                                         if job.status == job.StatusCodes.RUNNING]})


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = "cwl_execution/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context["query"] = query

        cwl_execution_results = set()
        job_results = set()

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            cwl_execution_results.update(Executable.objects.filter(
                (Q(project__owner=user) | Q(project__co_owner=user)) & Q(project__deleted__isnull=True) &
                (
                    Q(executable_id=num) |
                    Q(cwl_json__icontains=word) |
                    Q(type=num)
                )
            ))
            job_results.update(Job.objects.filter(
                (Q(project__owner=user) | Q(project__co_owner=user)) & Q(project__deleted__isnull=True) &
                (Q(job_id=num) |
                 Q(name__icontains=word) |
                 Q(status=num) |
                 Q(description__icontains=word) |
                 Q(output_text__icontains=word) |
                 Q(output_code=num) |
                 Q(cwl_json__icontains=word) |
                 Q(interval=num, repeat=True) |
                 Q(base_executable__executable_id=num) |
                 Q(setup_command__icontains=word))
            ))

        context["job_results"] = job_results
        context["cwl_execution_results"] = cwl_execution_results
        return context


class ManualPeriodicJobsView(HeliportLoginRequiredMixin, django_views.View):
    def get(self, request, **kwargs):
        result = tasks.run_periodic_jobs()
        return HttpResponse(f'OK. {result} jobs started')


##########################################
#               REST API                 #
##########################################

class ExecutableViewSet(viewsets.ModelViewSet):
    """Tools, Workflows and Arbitrary CWL"""
    serializer_class = ExecutableSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['executable_id', 'cwl_type', 'project', 'type']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return Executable.objects.filter(is_object_member(user))


class JobViewSet(viewsets.ModelViewSet):
    """
    Jobs.
    To execute a job go to its UserJobConfig like .../user_job_config/user_job_config id/execute/
    """
    serializer_class = JobSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['job_id', 'project', 'name', 'status', 'output_code', 'repeat',
                        'interval', 'base_executable', 'setup_command']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return Job.objects.filter(is_object_member(user))


class UserJobConfigViewSet(viewsets.ModelViewSet):
    """
    User specific config for a job.
    You can Execute your config by appending /execute/ to the URL. For example: .../user_job_config/123/execute/
    """
    serializer_class = UserJobConfigSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['user_job_config_id', 'job', 'ssh_login', 'directory']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return UserJobConfig.objects.filter(ssh_login__user=user)

    @action(detail=True, methods=['get'])
    def execute(self, request, pk=None):
        config = self.get_object()
        start_job(config, request, request.GET.get('debug', False))
        return RestResponse(f'job { config.job.pk } started')
