import json
from django.test import TestCase
from django.contrib.auth.models import User
from django.shortcuts import reverse

from heliport.models import Project, HeliportUser
from .models import DataManagementPlan


class DataManagementPlanTests(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username='testuser')
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.client.force_login(self.auth_user)
        self.project = Project.objects.create(owner=self.user)

    def test_list(self):
        dmp, is_new = DataManagementPlan.objects.get_or_create(project=self.project)

        response = self.client.get(reverse('data_management_plan:list', kwargs={'project': self.project.pk}))
        self.assertEquals(200, response.status_code)
        self.assertEquals(response.context_data['project'].datamanagementplan_set.get(pk=dmp.pk), dmp)

        response = self.client.get(reverse('data_management_plan:update', kwargs={'project': self.project.pk, 'pk': dmp.pk}))
        self.assertEquals(200, response.status_code)
        self.assertEquals(response.context_data['project'].datamanagementplan_set.get(pk=dmp.pk), dmp)

    def test_create(self):
        self.client.post(reverse('data_management_plan:list', kwargs={'project': self.project.pk}),
                         data={
                             'description': 'data_management321',
                             'link': 'https://www.google.com'
                         })
        test_obj = self.project.datamanagementplan_set.filter(description='data_management321').first()
        self.assertIsNotNone(test_obj)
        self.assertEquals('https://www.google.com', test_obj.link)

    def test_update(self):
        dmp, is_new = DataManagementPlan.objects.get_or_create(project=self.project,
                                                               description='data_management456',
                                                               link='http://www.google.com')
        response = self.client.post(reverse('data_management_plan:update', kwargs={'project': self.project.pk, 'pk': dmp.pk}),
                                    data={
                                        'description': 'data_plan789',
                                        'link': 'https://www.bing.com'
                                    }, follow=True)
        self.assertEquals(200, response.status_code)
        dmp = response.context_data['project'].datamanagementplan_set.get(pk=dmp.pk)
        self.assertEquals('data_plan789', dmp.description)
        self.assertEquals('https://www.bing.com', dmp.link)

    def test_delete(self):
        data_plan = DataManagementPlan.objects.create(project=self.project)
        data_plan2 = DataManagementPlan.objects.create(project=self.project)
        response = self.client.post(reverse('data_management_plan:list', kwargs={'project': self.project.pk}),
                                    data={
                                        'remove': data_plan2.pk
                                    }, follow=True)
        self.assertEquals(200, response.status_code)
        self.assertIsNotNone(self.project.datamanagementplan_set.filter(pk=data_plan.pk).first())
        self.assertIsNone(self.project.datamanagementplan_set.filter(pk=data_plan2.pk).first())

    def test_add_protocol_create(self):
        self.client.post(reverse('data_management_plan:list', kwargs={'project': self.project.pk}),
                         data={
                             'description': 'data_management321',
                             'link': 'www.google.com'
                         })
        test_obj = self.project.datamanagementplan_set.filter(description='data_management321').first()
        self.assertIsNotNone(test_obj)
        self.assertEquals('https://www.google.com', test_obj.link)

    def test_add_protocol_update(self):
        dmp, is_new = DataManagementPlan.objects.get_or_create(project=self.project,
                                                               description='data_management456',
                                                               link='http://www.google.com')
        response = self.client.post(
            reverse('data_management_plan:update', kwargs={'project': self.project.pk, 'pk': dmp.pk}),
            data={
                'description': 'data_plan789',
                'link': 'www.bing.com'
            }, follow=True)
        dmp = response.context_data['project'].datamanagementplan_set.get(pk=dmp.pk)
        self.assertEquals('https://www.bing.com', dmp.link)


class SearchAndAPITest(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username='testuser')
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.project, is_new = Project.objects.get_or_create(owner=self.user)
        self.client.force_login(self.auth_user)

    def test_access_search(self):
        response = self.client.get(reverse('data_management_plan:search'))
        self.assertEquals(200, response.status_code)

    def test_data_management_plan_findable(self):
        DataManagementPlan.objects.create(project=self.project, description='DMP321')
        response = self.client.get(f'{reverse("data_management_plan:search")}?q=321')
        self.assertContains(response, 'DMP321')

    def test_api(self):
        DataManagementPlan.objects.create(project=self.project, description='DMP654')
        response = self.client.get(reverse('data_management_plan:api_root'))
        self.assertEquals(200, response.status_code)
        urls = json.loads(response.content.decode())
        response = self.client.get(urls['data_management_plan'])
        self.assertEquals(200, response.status_code)
        self.assertContains(response, 'DMP654')
