from django.db.models import Q
from django.shortcuts import redirect, get_object_or_404, reverse
from rest_framework import viewsets
from rest_framework import permissions as rest_permissions

from .serializers import DataManagementPlanSerializer
from heliport.permissions import HeliportObjectPermission, is_object_member
from heliport.custom_views import HeliportObjectView, HeliportProjectView, HeliportLoginRequiredMixin, django_views
from .models import *


class DataManagementPlanView(HeliportProjectView, django_views.CreateView):
    model = DataManagementPlan
    template_name = "data_management_plan/data_management_plan.html"
    fields = ["description", "link"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        data_management_plan_id = request.POST.get("remove")
        if data_management_plan_id:
            data_management_plan = get_object_or_404(DataManagementPlan, pk=data_management_plan_id)
            data_management_plan.delete()
            return redirect(reverse("data_management_plan:list", kwargs=self.kwargs))
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        form.instance.project = project
        if not form.instance.link.startswith("http://") and not form.instance.link.startswith("https://"):
            form.instance.link = "https://" + form.instance.link
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_management_plan:list", kwargs=self.kwargs)


class DataManagementPlanUpdateView(HeliportObjectView, django_views.UpdateView):
    model = DataManagementPlan
    template_name = "data_management_plan/data_management_plan.html"
    fields = ["description", "link"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["update"] = True
        return context

    def form_valid(self, form):
        if not form.instance.link.startswith("http://") and not form.instance.link.startswith("https://"):
            form.instance.link = "https://" + form.instance.link
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_management_plan:list", kwargs={"project": self.get_object().project.pk})


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = "data_management_plan/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context["query"] = query

        data_management_plan_results = set()

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            data_management_plan_results.update(DataManagementPlan.objects.filter(
                (Q(project__owner=user) | Q(project__co_owner=user)) & Q(project__deleted__isnull=True) &
                (
                    Q(data_management_plan_id=num) |
                    Q(description__icontains=word) |
                    Q(link__icontains=word)
                )
            ))

        context["data_management_plan_results"] = data_management_plan_results
        return context


##########################################
#               REST API                 #
##########################################

class DataManagementPlanViewSet(viewsets.ModelViewSet):
    """Data Management Plan"""
    serializer_class = DataManagementPlanSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['data_management_plan_id', 'link', 'project']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return DataManagementPlan.objects.filter(is_object_member(user))
