from django.db.models import Q
from django.shortcuts import redirect, get_object_or_404, reverse
from rest_framework import viewsets
from rest_framework import permissions as rest_permissions

from .serializers import DataSourceSerializer
from heliport.permissions import HeliportObjectPermission, is_object_member
from heliport.custom_views import HeliportObjectView, HeliportProjectView, HeliportLoginRequiredMixin, django_views
from .models import *


class DataSourceView(HeliportProjectView, django_views.CreateView):
    model = DataSource
    template_name = "data_source/data_source.html"
    fields = ["uri", "description", "name", "login"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = context["project"]
        context["sources"] = [{"data_source_id": source.data_source_id,
                               "protocol": source.uri.split(":")[0],
                               "name": source.name,
                               "description": source.description,
                               "uri": source.uri,
                               "owner": (source.login is None) or (self.request.user.heliportuser == source.login.user)}
                              for source in project.datasource_set.all()]
        user = self.request.user.heliportuser
        context["protocols"] = DataSource.PROTOCOL_CHOICES
        context["logins"] = user.logininfo_set.all()
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        data_source_id = request.POST.get("remove")
        if data_source_id:
            data_source = get_object_or_404(DataSource, pk=data_source_id)
            data_source.delete()
            return redirect(reverse("data_source:list", kwargs=self.kwargs))
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        form.instance.project = project
        protocol = self.request.POST["protocol"]
        server_adress = self.request.POST["server_adress"]

        if protocol == "http" and not form.instance.uri.startswith("http://"):
            form.instance.uri = "http://" + form.instance.uri
        elif protocol == "https" and not form.instance.uri.startswith("https://"):
            form.instance.uri = "https://" + form.instance.uri
        elif protocol == "file" and not form.instance.uri.startswith("file://"):
            form.instance.uri = "file://" + form.instance.uri
        elif protocol == "ftp":
            if not form.instance.uri.startswith("/") and not server_adress.endswith("/"):
                form.instance.uri = "/" + form.instance.uri
            form.instance.uri = server_adress + form.instance.uri
            if not form.instance.uri.startswith("ftp://"):
                form.instance.uri = "ftp://" + form.instance.uri
        elif protocol == "ftps":
            if not form.instance.uri.startswith("/") and not server_adress.endswith("/"):
                form.instance.uri = "/" + form.instance.uri
            form.instance.uri = server_adress + form.instance.uri
            if not form.instance.uri.startswith("ftps://"):
                form.instance.uri = "ftps://" + form.instance.uri
        elif protocol == "ssh":
            form.instance.uri = "ssh://" + form.instance.uri
        else:
            if "://" not in form.instance.uri:
                form.instance.uri = "://" + form.instance.uri

        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_source:list", kwargs=self.kwargs)


class DataSourceUpdateView(HeliportObjectView, django_views.UpdateView):
    model = DataSource
    template_name = "data_source/data_source.html"
    fields = ["uri", "description", "name", "login"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["sources"] = [{"data_source_id": source.data_source_id,
                               "protocol": source.uri.split(":")[0],
                               "name": source.name,
                               "description": source.description,
                               "uri": source.uri,
                               "owner": (source.login is None) or (self.request.user.heliportuser == source.login.user)}
                              for source in self.get_object().project.datasource_set.all()]
        protocol = self.request.GET.get("protocol", self.object.uri.split("://")[0])
        user = self.request.user.heliportuser

        if protocol == ("ftp" or "ftps"):
            context["server_adress"] = self.object.uri.split("://", 1)[1].split("/", 1)[0]
            context["uri"] = self.object.uri.split("://", 1)[1].split("/", 1)[1]
        else:
            context["uri"] = self.object.uri.split("://", 1)[1]

        context["protocol"] = protocol
        context["protocols"] = DataSource.PROTOCOL_CHOICES
        context["logins"] = user.logininfo_set.all()
        context["login"] = self.object.login
        context["update"] = True
        return context

    def form_valid(self, form):
        if (self.object.login is None) or (self.request.user.heliportuser == self.object.login.user):
            protocol = self.request.POST["protocol"]
            server_adress = self.request.POST["server_adress"]

            if protocol == "http" and not form.instance.uri.startswith("http://"):
                form.instance.uri = "http://" + form.instance.uri
            elif protocol == "https" and not form.instance.uri.startswith("https://"):
                form.instance.uri = "https://" + form.instance.uri
            elif protocol == "file" and not form.instance.uri.startswith("file://"):
                form.instance.uri = "file://" + form.instance.uri
            elif protocol == "ftp":
                if not form.instance.uri.startswith("/") and not server_adress.endswith("/"):
                    form.instance.uri = "/" + form.instance.uri
                form.instance.uri = server_adress + form.instance.uri
                if not form.instance.uri.startswith("ftp://"):
                    form.instance.uri = "ftp://" + form.instance.uri
            elif protocol == "ftps":
                if not form.instance.uri.startswith("/") and not server_adress.endswith("/"):
                    form.instance.uri = "/" + form.instance.uri
                form.instance.uri = server_adress + form.instance.uri
                if not form.instance.uri.startswith("ftps://"):
                    form.instance.uri = "ftps://" + form.instance.uri
            elif protocol == "ssh":
                form.instance.uri = "ssh://" + form.instance.uri
            else:
                if "://" not in form.instance.uri:
                    form.instance.uri = "://" + form.instance.uri

        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_source:list", kwargs={"project": self.get_object().project.pk})


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = "data_source/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context["query"] = query

        data_source_results = set()

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            data_source_results.update(DataSource.objects.filter(
                (Q(project__owner=user) | Q(project__co_owner=user)) & Q(project__deleted__isnull=True) &
                (Q(data_source_id=num) |
                 Q(uri__icontains=word) |
                 Q(description__icontains=word) |
                 Q(name__icontains=word) |
                 Q(login__machine__icontains=word) |
                 Q(login__username__icontains=word) |
                 Q(project__project_name__icontains=word))
            ))

        context["data_source_results"] = data_source_results
        return context


##########################################
#               REST API                 #
##########################################

class DataSourceViewSet(viewsets.ModelViewSet):
    """Data Source"""
    serializer_class = DataSourceSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['data_source_id', 'name', 'uri', 'project']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return DataSource.objects.filter(is_object_member(user))
