from collections import Counter
from django.shortcuts import get_object_or_404, redirect, reverse
from django.db.models import Q
from django.http.response import JsonResponse
from django.conf import settings
from rest_framework import viewsets
from rest_framework import permissions as rest_permissions
from heliport.custom_views import HeliportObjectView, HeliportProjectView, HeliportLoginRequiredMixin, django_views
from heliport.models import Project
from heliport.utils import handles
from heliport.permissions import HeliportObjectPermission, is_object_member
from .models import DigitalObject
from .serializers import DigitalObjectSerializer


class ObjectsView(HeliportProjectView, django_views.TemplateView):
    template_name = 'digital_objects/digital_objects.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = context['project']
        edit_pk = None
        try:
            edit_pk = int(self.request.GET.get('edit'))
        except (TypeError, ValueError):
            pass
        try:
            context['highlight'] = int(self.request.GET.get('highlight'))
        except (TypeError, ValueError):
            pass
        if edit_pk is not None:
            context['digital_object'] = get_object_or_404(DigitalObject, pk=edit_pk)
        context['update'] = edit_pk is not None
        context['digital_objects'] = project.digitalobject_set.all()
        return context

    def post(self, request, *args, **kwargs):
        relation = request.POST.get('relation', '')
        category = request.POST.get('category', '')
        description = request.POST.get('description', '')
        object_pk = request.POST.get('object_pk')
        reference_handle = request.POST.get('referenced_handle', '')
        use_handle = request.POST.get('handle')
        handle_prefix = 'hdl.handle.net/'
        if reference_handle.startswith(handle_prefix):
            reference_handle = reference_handle[len(handle_prefix):]
        project = get_object_or_404(Project, pk=self.kwargs['project'])
        new_object = DigitalObject()
        if object_pk:
            new_object = get_object_or_404(DigitalObject, pk=object_pk)
        elif use_handle:
            new_object.handle = reference_handle
        new_object.relation = relation
        new_object.category = category
        new_object.description = description
        new_object.project = project
        new_object.save()

        if settings.REGISTER_HANDLES and not new_object.handle:
            handle_name = f'HELIPORT_OBJ_{new_object.pk}'
            url = str(reverse('digital_objects:list', kwargs={'project': project.pk})) + f'?highlight={new_object.pk}'
            full_url = f'{request.scheme}://{request.get_host()}{url}'
            new_object.handle = handles.register_json_handle(handle_name, category, full_url).json().get('handle')
            new_object.save()

        return redirect('digital_objects:list', project=project.pk)


class QueryHandleView(django_views.View):
    def get(self, request, *args, **kwargs):
        handle = request.GET.get('handle')
        https_prefix = 'https://'
        http_prefix = 'http://'
        server_prefix = 'hdl.handle.net/'
        if handle.startswith(https_prefix):
            handle = handle[len(https_prefix):]
        if handle.startswith(http_prefix):
            handle = handle[len(http_prefix):]
        if handle.startswith(server_prefix):
            handle = handle[len(server_prefix):]

        values = [
            {
                'project': obj.project.hzdr_id,
                'relation': obj.relation,
                'category': obj.category,
                'description': obj.description
            }
            for obj in DigitalObject.objects.filter(handle=handle)
        ]

        return JsonResponse({
            'handle': handle,
            'found_match': bool(values),
            'project': self.find_most_often(values, 'project'),
            'relation': self.find_most_often(values, 'relation'),
            'category': self.find_most_often(values, 'category'),
            'description': self.find_most_often(values, 'description')
        })

    @staticmethod
    def find_most_often(values, attribute):
        hist = Counter(v[attribute] for v in values)
        sorted_hist = sorted((count, value) for value, count in hist.most_common())
        result = None
        if len(sorted_hist) > 0:
            count, result = sorted_hist[0]
        return result


class RelationAutocompleteView(django_views.View):
    def get(self, request, *args, **kwargs):
        pattern = request.GET.get('q', '')
        relation_values = DigitalObject.objects.filter(relation__icontains=pattern).values_list('relation', flat=True)
        hist = Counter(relation_values)
        result = [relation for relation, count in hist.most_common(10)]
        return JsonResponse({'query': pattern, 'results': result})


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = "digital_objects/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context["query"] = query

        digital_objects_results = set()

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            digital_objects_results.update(DigitalObject.objects.filter(
                (Q(project__owner=user) | Q(project__co_owner=user)) & Q(project__deleted__isnull=True) &
                (
                    Q(digital_object_id=num) |
                    Q(description__icontains=word) |
                    Q(handle__icontains=word) |
                    Q(category__icontains=word) |
                    Q(relation__icontains=word)
                )
            ))

        context["digital_object_results"] = digital_objects_results
        return context


##########################################
#               REST API                 #
##########################################

class DigitalObjectViewSet(viewsets.ModelViewSet):
    """Digital Objects"""
    serializer_class = DigitalObjectSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['digital_object_id', 'project', 'handle', 'relation', 'category']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return DigitalObject.objects.filter(is_object_member(user))
