import json
from django.test import TestCase
from django.contrib.auth.models import User
from django.shortcuts import reverse

from heliport.models import Project, HeliportUser
from .models import Documentation


class DocumentationTests(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username='testuser')
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.client.force_login(self.auth_user)
        self.project = Project.objects.create(owner=self.user)

    def test_list(self):
        doc, is_new = Documentation.objects.get_or_create(project=self.project, system=1)

        response = self.client.get(reverse('documentation:list', kwargs={'project': self.project.pk}))
        self.assertEquals(200, response.status_code)
        self.assertEquals(response.context_data['project'].documentation_set.get(pk=doc.pk), doc)

        response = self.client.get(reverse('documentation:update', kwargs={'project': self.project.pk, 'pk': doc.pk}))
        self.assertEquals(200, response.status_code)
        self.assertEquals(response.context_data['project'].documentation_set.get(pk=doc.pk), doc)

    def test_create(self):
        self.client.post(reverse('documentation:list', kwargs={'project': self.project.pk}),
                         data={
                             'description': 'docu321',
                             'link': 'https://www.google.com',
                             'system': 1
                         })
        test_obj = self.project.documentation_set.filter(description='docu321').first()
        self.assertIsNotNone(test_obj)
        self.assertEquals('https://www.google.com', test_obj.link)

    def test_update(self):
        doc, is_new = Documentation.objects.get_or_create(project=self.project,
                                                          description='docu456',
                                                          link='http://www.google.com',
                                                          system=1)
        response = self.client.post(reverse('documentation:update', kwargs={'project': self.project.pk, 'pk': doc.pk}),
                                    data={
                                        'description': 'docu789',
                                        'link': 'https://www.bing.com',
                                        'system': 2
                                    }, follow=True)
        self.assertEquals(200, response.status_code)
        doc = response.context_data['project'].documentation_set.get(pk=doc.pk)
        self.assertEquals('docu789', doc.description)
        self.assertEquals('https://www.bing.com', doc.link)
        self.assertEquals(2, doc.system)

    def test_delete(self):
        doc = Documentation.objects.create(project=self.project, system=2)
        doc2 = Documentation.objects.create(project=self.project, system=2)
        response = self.client.post(reverse('documentation:list', kwargs={'project': self.project.pk}),
                                    data={
                                        'remove': doc2.pk
                                    }, follow=True)
        self.assertEquals(200, response.status_code)
        self.assertIsNotNone(self.project.documentation_set.filter(pk=doc.pk).first())
        self.assertIsNone(self.project.documentation_set.filter(pk=doc2.pk).first())

    def test_add_protocol_create(self):
        self.client.post(reverse('documentation:list', kwargs={'project': self.project.pk}),
                         data={
                             'description': 'docu321',
                             'link': 'www.google.com',
                             'system': 1
                         })
        test_obj = self.project.documentation_set.filter(description='docu321').first()
        self.assertIsNotNone(test_obj)
        self.assertEquals('https://www.google.com', test_obj.link)

    def test_add_protocol_update(self):
        doc, is_new = Documentation.objects.get_or_create(project=self.project,
                                                          description='docu456',
                                                          link='http://www.google.com',
                                                          system=1)
        response = self.client.post(reverse('documentation:update', kwargs={'project': self.project.pk, 'pk': doc.pk}),
                                    data={
                                        'description': 'docu789',
                                        'link': 'www.bing.com',
                                        'system': 2
                                    }, follow=True)
        self.assertEquals(200, response.status_code)
        doc = response.context_data['project'].documentation_set.get(pk=doc.pk)
        self.assertEquals('https://www.bing.com', doc.link)


class SearchAndAPITest(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username='testuser')
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.project, is_new = Project.objects.get_or_create(owner=self.user)
        self.client.force_login(self.auth_user)

    def test_access_search(self):
        response = self.client.get(reverse('documentation:search'))
        self.assertEquals(200, response.status_code)

    def test_documentation_findable(self):
        Documentation.objects.create(project=self.project, description='DOCUMENTATION321', system=1)
        response = self.client.get(f'{reverse("documentation:search")}?q=321')
        self.assertContains(response, 'DOCUMENTATION321')

    def test_api(self):
        Documentation.objects.create(project=self.project, description='DOCU654', system=2)
        response = self.client.get(reverse('documentation:api_root'))
        self.assertEquals(200, response.status_code)
        urls = json.loads(response.content.decode())
        response = self.client.get(urls['documentation'])
        self.assertEquals(200, response.status_code)
        self.assertContains(response, 'DOCU654')
