from django.db.models import Q
from django.shortcuts import redirect, get_object_or_404, reverse
from rest_framework import viewsets
from rest_framework import permissions as rest_permissions

from .serializers import DocumentationSerializer
from heliport.permissions import HeliportObjectPermission, is_object_member
from heliport.custom_views import HeliportObjectView, HeliportProjectView, HeliportLoginRequiredMixin, django_views
from .models import *


class DocumentationView(HeliportProjectView, django_views.CreateView):
    model = Documentation
    template_name = "documentation/documentation.html"
    fields = ["description", "link", "system"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["systems"] = Documentation.DOCU_SYSTEMS
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        documentation_id = request.POST.get("remove")
        if documentation_id:
            documentation = get_object_or_404(Documentation, pk=documentation_id)
            documentation.delete()
            return redirect(reverse("documentation:list", kwargs=self.kwargs))
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        form.instance.project = project
        if not form.instance.link.startswith("http://") and not form.instance.link.startswith("https://"):
            form.instance.link = "https://" + form.instance.link
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("documentation:list", kwargs=self.kwargs)


class DocumentationUpdateView(HeliportObjectView, django_views.UpdateView):
    model = Documentation
    template_name = "documentation/documentation.html"
    fields = ["description", "link", "system"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["systems"] = Documentation.DOCU_SYSTEMS
        context["update"] = True
        return context

    def form_valid(self, form):
        if not form.instance.link.startswith("http://") and not form.instance.link.startswith("https://"):
            form.instance.link = "https://" + form.instance.link
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("documentation:list", kwargs={'project': self.object.project.pk})


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = "documentation/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context["query"] = query

        documentation_results = set()

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            documentation_results.update(Documentation.objects.filter(
                (Q(project__owner=user) | Q(project__co_owner=user)) & Q(project__deleted__isnull=True) &
                (
                    Q(documentation_id=num) |
                    Q(description__icontains=word) |
                    Q(system=num) |
                    Q(link__icontains=word)
                )
            ))

        context["documentation_results"] = documentation_results
        return context


##########################################
#               REST API                 #
##########################################

class DocumentationViewSet(viewsets.ModelViewSet):
    """Documentation"""
    serializer_class = DocumentationSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['documentation_id', 'system', 'project']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return Documentation.objects.filter(is_object_member(user))
