import abc
import environ

from heliport.logger import logger

from utils import data_command_line_execution
import typing
from .ldap_user_management import LDAPUser
from .gate_user_management import GateUser

env = environ.Env()


class GateProject:
    def __init__(self, main_dict, user_dict):
        self.gate_id = main_dict["id"]
        self.title = main_dict["title"]
        self.abstract = main_dict["abstract"]
        self.proposal = main_dict["proposalnumber"]
        self.restricted = main_dict["restricted"] == "true"
        self.status = main_dict.get("status", "no status")
        self.proposer = user_dict["proposer"]
        self.responsible_experimentalist = user_dict["responsible_experimentalist"]
        self.co_proposers = user_dict["coproposers"]
        self.local_contacts = user_dict["local_contacts"]


class GateQueries:
    @staticmethod
    @abc.abstractmethod
    def create_or_update_gate_project(gate_project: GateProject):
        pass


class ProposalManager:
    connection: typing.Optional[data_command_line_execution.CommandExecutor]

    def __init__(self):
        self.connection = None

    def __enter__(self):
        self.connection = data_command_line_execution.create_data_command_executor()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.connection.close()

    @abc.abstractmethod
    def download_proposals(self):
        pass

    @abc.abstractmethod
    def save_proposal(self, proposal, gate_queries: GateQueries):
        pass

    @abc.abstractmethod
    def import_users(self, projects, user_queries):
        pass


class HZDRGateProposalManager(ProposalManager):
    def __init__(self):
        super().__init__()
        self._token = None

        self.gate_project_url = "https://www.hzdr.de/db/proposalmanagement/proposals/views"
        self.user_relation_url = "https://www.hzdr.de/db/proposalmanagement/proposals/users"
        self.token_url = "https://www.hzdr.de/db/proposalmanagement/oauth/token"
        self.gate_user = env('GATE_TOKEN_STR')

    def retrieve_token(self):
        # example data:
        # {"access_token": "V5prXY7o8DidohGe02DXnw", "token_type": "bearer", "expires_in": 3600}
        logger.info("=== GET BEARER TOKEN ===")
        command = f'curl -i -k --user {self.gate_user} --data "grant_type=client_credentials" {self.token_url}'
        data = self.connection.execute(command).extract_http_json()

        logger.debug(data["access_token"])
        self._token = data["access_token"]

    @property
    def token(self):
        if self._token is None:
            self.retrieve_token()
        return self._token

    def get_paged_data(self, initial_url, object_name):
        # example data:
        # {
        #    "items": [... ...],
        #    "hasMore": true, "limit": 50, "offset": 0, "count": 50, "links": [
        #        ...
        #        {
        #            "rel": "next",
        #            "href": "https://www.hzdr.de/db/proposalmanagement/proposals/views?offset=50"
        #        }
        #    ]
        # }
        logger.info("=== GET DATA ===")
        url = initial_url
        has_more = True
        all_items = []

        while has_more:
            command = f'curl -i -k -H "Authorization: Bearer {self.token}" {url}'
            data = self.connection.execute(command).extract_http_json()
            has_more = data["hasMore"]
            all_items.extend(data["items"])

            next_urls = [url["href"] for url in data["links"] if url["rel"] == "next"]
            if has_more:
                url = next_urls[0]

            logger.debug(f"loaded {object_name}: {data['offset'] + data['count']}")

        return all_items

    def insert_user_objects(self, users):
        # example users
        # {
        #    'id': 1789,
        #    'proposer': 2000123,
        #    'coproposers': [{'id': 2000456}],
        #    'responsible_experimentalist': 2000123,
        #    'local_contacts': [{'id': 101234}]
        # }
        if "coproposers" not in users:
            logger.error(f"user relation has no coproposers: {users}")
        if "local_contacts" not in users:
            logger.error(f"user relation has no local_contacts: {users}")

        users["proposer"] = GateUser(users["proposer"], self)
        users["responsible_experimentalist"] = GateUser(users["responsible_experimentalist"], self)
        users["coproposers"] = [GateUser(user["id"], self) for user in users["coproposers"]]
        users["local_contacts"] = [LDAPUser(user["id"]) for user in users["local_contacts"]]

    def download_proposals(self):
        gate_project_dicts = self.get_paged_data(self.gate_project_url, "gate_projects")
        user_relation = self.get_paged_data(self.user_relation_url, "user_relations")

        flawed_data_count = sum("coproposers" not in users for users in user_relation)
        if flawed_data_count:
            logger.error(f"{flawed_data_count} user-relations have no co-proposers")

        user_index = {gate_project["id"]: gate_project for gate_project in user_relation}

        gate_projects = []
        for project in gate_project_dicts:
            users = user_index[project["id"]]
            self.insert_user_objects(users)
            gate_projects.append(GateProject(project, users))

        return gate_projects

    def save_proposal(self, gate_project, gate_queries):
        gate_queries.create_or_update_gate_project(gate_project)

    def import_users(self, projects, user_queries):
        GateUser.import_users(projects, user_queries, self.connection)
        LDAPUser.import_users(projects, user_queries, self.connection)

    def number_of_proposals_changed(self, last_count):
        auth = f'"Authorization: Bearer {self.token} "'
        command = f'curl -i -k -H {auth} {self.user_relation_url}?offset={last_count - 1}'
        data = self.connection.execute(command).extract_http_json()
        return data["count"] != 1 or data["hasMore"]
