from django.shortcuts import redirect, get_object_or_404, reverse, HttpResponse
from django.db.models import Q
from rest_framework import viewsets
from rest_framework import permissions as rest_permissions

from .serializers import GateProjectSerializer, GateUserSerializer
from heliport.permissions import HeliportObjectPermission, is_object_member, StaffOrReadOnly
from heliport.custom_views import HeliportProjectView, HeliportLoginRequiredMixin, django_views
from .models import *
from . import tasks


class ConnectToGateProjectView(HeliportProjectView, django_views.ListView):
    template_name = "gate_connection/gate_connect.html"
    model = GateProject
    context_object_name = "gate_project_list"

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context["query"] = self.request.GET.get("q", "")
        return context

    def get_queryset(self):
        user = self.request.user.heliportuser
        proposed = user.gateproject_set.all()
        co_proposed = user.co_proposed_gate_projects.all()
        user_gate_projects = set(proposed.union(co_proposed))

        query = self.request.GET.get("q")
        if query:
            for word in query.split():
                num = int(word) if word.isnumeric() else -1
                user_gate_projects.intersection_update(
                    GateProject.objects.filter((
                        Q(gate_id=num) |
                        Q(title__icontains=word) |
                        Q(proposer__display_name__icontains=word) |
                        Q(co_proposer__display_name__icontains=word) |
                        Q(abstract__icontains=word) |
                        Q(proposal__icontains=word) |
                        Q(responsible_experimentalist__display_name__icontains=word) |
                        Q(local_contacts__display_name__icontains=word)) &
                        (Q(proposer=user) | Q(co_proposer=user))
                    )
                )

        return user_gate_projects

    def post(self, request, *args, **kwargs):
        gate_id = request.POST.get("connect")
        gate_project = get_object_or_404(GateProject, pk=gate_id)
        project = get_object_or_404(Project, pk=kwargs["project"])

        if Project.objects.filter(gateproject=gate_project):
            return redirect(reverse("gate_connection:connect", kwargs=kwargs))

        if project.gateproject_set.first() is not None:
            return redirect(reverse("gate_connection:connect", kwargs=kwargs))

        gate_project.project = project
        if not project.project_name:
            project.project_name = gate_project.title

        gate_project.save()
        project.save()

        # add co_proposers to owners
        project.co_owner.add(*gate_project.co_proposer.exclude(pk=project.owner.pk))

        # add proposer to owners
        if project.owner != gate_project.proposer:
            project.co_owner.add(gate_project.proposer)

        return redirect(reverse("gate_connection:connected", kwargs={"project": project.pk, "pk": gate_project.pk}))


class ConnectedToGateProjectView(HeliportProjectView, django_views.DetailView):
    template_name = "gate_connection/gate_connected.html"
    model = Project
    pk_url_kwarg = 'project'

    def get(self, request, *args, **kwargs):
        if not self.get_object().gateproject_set.all():
            return redirect(reverse("gate_connection:connect", kwargs={"project": self.get_object().pk}))
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["is_owner"] = self.request.user.heliportuser == self.object.owner
        context["gate_project"] = self.get_object().gateproject_set.first()
        proposers = {*self.object.gateproject_set.first().co_proposer.all(), self.object.gateproject_set.first().proposer}
        context["import"] = bool(proposers - self.object.members())
        return context

    def post(self, request, *args, **kwargs):
        action = request.POST.get("action")
        project = get_object_or_404(Project, pk=kwargs["project"])
        gate_project = project.gateproject_set.first()

        if action == "import":
            # add co_proposers to owners
            project.co_owner.add(*gate_project.co_proposer.exclude(pk=project.owner.pk))

            # add proposer to owners
            if project.owner != gate_project.proposer:
                project.co_owner.add(gate_project.proposer)

            return redirect(reverse("gate_connection:connected", kwargs=kwargs))

        if action == "import_name":
            project.project_name = gate_project.title
            project.save()
            return redirect(reverse("gate_connection:connected", kwargs=kwargs))

        if action == "disconnect":
            project.gateproject_set.clear()
            project.save()

            return redirect(reverse("gate_connection:connect", kwargs={"project": project.pk}))


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = "gate_connection/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context["query"] = query

        gate_connection_results = set()

        if self.request.user.is_staff:
            gate_projects = GateProject.objects
        else:
            gate_projects = GateProject.objects.filter(
                Q(proposer=user) |
                Q(co_proposer=user) |
                Q(responsible_experimentalist=user) |
                Q(local_contacts=user)
            )

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            gate_connection_results.update(gate_projects.filter(
                Q(gate_id=num) |
                Q(title__icontains=word) |
                Q(proposer__display_name__icontains=word) |
                Q(co_proposer__display_name__icontains=word) |
                Q(abstract__icontains=word) |
                Q(responsible_experimentalist__display_name__icontains=word) |
                Q(local_contacts__display_name__icontains=word)
            ))

        context["gate_connection_results"] = gate_connection_results
        return context


class ManualGateDownloadView(HeliportLoginRequiredMixin, django_views.View):
    def get(self, request, **kwargs):
        tasks.update_gate_projects()
        return HttpResponse('OK')


##########################################
#               REST API                 #
##########################################

class GateProjectViewSet(viewsets.ModelViewSet):
    """Gate Project"""
    serializer_class = GateProjectSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['gate_id', 'title', 'proposer', 'co_proposer', 'proposal', 'restricted',
                        'responsible_experimentalist', 'local_contacts', 'project', 'status']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return GateProject.objects.filter(is_object_member(user))


class GateUserViewSet(viewsets.ModelViewSet):
    """Gate User"""
    serializer_class = GateUserSerializer
    permission_classes = [rest_permissions.IsAuthenticated, StaffOrReadOnly]
    filterset_fields = ['gate_user_id', 'gate_id', 'heliport_user']

    def get_queryset(self):
        return GateUser.objects.all()
