from django.db import models
from django.contrib.auth.models import User
from utils.command_line_execution import create_command_executor
from utils.IDs import generate_object_id
from heliport.utils import environment
import environ

env = environ.Env()


class Group(models.Model):
    group_id = models.CharField(primary_key=True, max_length=100)

    def __str__(self):
        return str(self.group_id)


class HeliportUser(models.Model):
    user_id = models.AutoField(primary_key=True)
    authentication_backend_id = models.CharField(max_length=100, null=True)
    auth_user = models.OneToOneField(User, on_delete=models.SET_NULL, null=True, blank=True)
    display_name = models.CharField(max_length=100, default="<User>")
    orcid = models.CharField(max_length=100, null=True, blank=True)

    def __str__(self):
        return str(self.display_name)


class Project(models.Model):
    project_id = models.AutoField(primary_key=True)
    hzdr_id = models.CharField(max_length=100, unique=True)
    uuid = models.UUIDField(primary_key=False, default=generate_object_id, editable=False)
    handle = models.CharField(max_length=100, null=True, blank=True)
    project_name = models.CharField(max_length=100)
    owner = models.ForeignKey(HeliportUser, on_delete=models.SET_NULL, null=True)
    co_owner = models.ManyToManyField(HeliportUser, related_name="co_owned_projects")
    created = models.DateTimeField('date created', auto_now_add=True)
    deleted = models.DateTimeField('date deleted', null=True, blank=True)
    description = models.TextField(blank=True, default="")
    group = models.ForeignKey(Group, on_delete=models.SET_NULL, null=True, blank=True)

    def __str__(self):
        return str(self.project_name)

    def members(self):
        return {*self.co_owner.all(), self.owner}


class LoginInfo(models.Model):
    class LoginTypes(models.TextChoices):
        SSH = "ssh connection"
        USER_PASSWORD = "username and password"
        TOKEN = "authentication token"

    login_info_id = models.AutoField(primary_key=True)
    user = models.ForeignKey(HeliportUser, on_delete=models.CASCADE)
    username = models.CharField(max_length=100, blank=True)
    password = models.CharField(max_length=100, blank=True)
    key = models.TextField(blank=True)
    machine = models.CharField(max_length=100, blank=True)
    name = models.CharField(max_length=100)
    type = models.CharField(max_length=50, choices=LoginTypes.choices)
    data = models.TextField(blank=True)
    via = models.ForeignKey("self", on_delete=models.SET_NULL, null=True, blank=True)
    slurm = models.BooleanField(default=False)

    def __str__(self):
        return self.name

    def build_command_executor(self, depth=10):
        if self.type != self.LoginTypes.SSH:
            raise TypeError('not a ssh login')
        if depth <= 0:
            raise ValueError('max depth reached: may be caused by cyclic ssh login')

        indirection = None
        if self.via is not None:
            indirection = self.via.build_command_executor(depth=depth - 1)

        return create_command_executor(
            host=self.machine,
            username=self.username,
            key=self.key,
            key_password_callback=environment.get_paramiko_krypto_key,
            indirection=indirection,
            auto_connect=False
        )

    def build_command_executor_indirection(self):
        if self.via is None:
            return None

        return self.via.build_command_executor()

    @property
    def connected(self) -> bool:
        if self.type != self.LoginTypes.SSH:
            return False
        else:
            return bool(self.data) and bool(self.key)

    def get_to_connect_list(self, depth=10, exclude_self=False) -> list:
        if self.type != self.LoginTypes.SSH:
            raise TypeError('not a ssh login')
        if depth <= 0:
            raise ValueError('max depth reached: may be caused by cyclic ssh login')

        result = []
        if self.via is not None:
            result = self.via.get_to_connect_list(depth=depth-1)
        if not self.connected and not exclude_self:
            result.append(self)
        return result

    def get_unconnected_parents(self) -> list:
        return self.get_to_connect_list(exclude_self=True)
