import abc

from django_auth_ldap.backend import LDAPBackend
from django.contrib.auth import login
from django.conf import settings

from heliport.models import Group, HeliportUser
from heliport.logger import logger
from utils.data_command_line_execution import create_data_command_executor


class UserInformationManager:
    @abc.abstractmethod
    def update(self, user):
        pass

    @abc.abstractmethod
    def get_display_name(self, user):
        pass

    @abc.abstractmethod
    def get_authentication_backand_id(self, user):
        pass

    @abc.abstractmethod
    def login(self, user, request):
        pass

    @abc.abstractmethod
    def import_groups(self):
        pass

    @abc.abstractmethod
    def import_users_and_groups(self):
        pass

    @abc.abstractmethod
    def get_users_by_group(self, group):
        pass

    @abc.abstractmethod
    def fetch_user(self, user):
        pass


class LdapUserInformationManager(UserInformationManager):
    @staticmethod
    def get_users_by_command(command):
        prefix_cn = 'cn: '
        prefix_name = 'displayName: '
        with create_data_command_executor() as executor:
            users_str = executor.execute(command).stdout
        for user in users_str.split('\n\n'):
            cn = None
            name = None
            for line in user.splitlines():
                if line.startswith(prefix_cn):
                    cn = line[len(prefix_cn):]
                if line.startswith(prefix_name):
                    name = line[len(prefix_name):]
            if cn and name:
                yield name, cn

    def get_users_by_group(self, group, updated_users_byref=None):
        for name, cn in self.get_users_by_command(settings.GET_USERS_BY_GROUP.format(group=group)):
            db_user = HeliportUser.objects.filter(authentication_backend_id=cn).first()
            if db_user is None:
                db_user = HeliportUser.objects.filter(display_name=name).first()
            if db_user is None:
                db_user = HeliportUser()

            if db_user.display_name != name or db_user.authentication_backend_id != cn:
                db_user.display_name = name
                db_user.authentication_backend_id = cn
                db_user.save()
                if updated_users_byref is not None:
                    updated_users_byref.append(db_user)

            yield db_user

    def import_groups(self):
        with create_data_command_executor() as executor:
            groups = executor.execute(settings.GET_GROUPS_COMMAND).stdout
        new_group_count = 0
        for group in groups.splitlines():
            db_group, is_new = Group.objects.get_or_create(group_id=group)
            if is_new:
                new_group_count += 1
        logger.info(f'{new_group_count} new groups imported')
        return new_group_count

    def import_users_and_groups(self):
        self.import_groups()
        new_user_count = 0
        groups = list(Group.objects.all())
        for i, group in enumerate(groups):
            new_users = []
            total = len(list(self.get_users_by_group(group.group_id, new_users)))
            new_user_count += len(new_users)
            logger.debug(f'imported users of group {i:3}/{len(groups)} ({group} -> {len(new_users)}/{total} new)')
        logger.info(f'{new_user_count} new users imported or updated')
        return new_user_count

    def update(self, user):
        LDAPBackend().populate_user(str(user))

    def get_authentication_backand_id(self, user):
        return user.ldap_user.attrs["cn"][0]

    def get_display_name(self, user):
        return user.ldap_user.attrs["displayName"][0]

    def login(self, user, request):
        login(request, user, backend='django_auth_ldap.backend.LDAPBackend')

    def fetch_user(self, user):
        self.update(user)
        display_name = self.get_display_name(user)
        authentication_backend_id = self.get_authentication_backand_id(user)
        hp_user = HeliportUser.objects.filter(authentication_backend_id=authentication_backend_id).first()
        if hp_user is None:
            hp_user, user_is_new = HeliportUser.objects.get_or_create(auth_user=user)
        if (
                hp_user.display_name != display_name or
                hp_user.authentication_backend_id != authentication_backend_id or
                hp_user.auth_user != user
        ):
            hp_user.display_name = display_name
            hp_user.authentication_backend_id = authentication_backend_id
            hp_user.auth_user = user
            hp_user.save()


authentication_backand_manager = LdapUserInformationManager()
