from django.http import JsonResponse
from django.shortcuts import render, redirect, reverse, get_object_or_404, HttpResponse
from django.db.models import Q
from django.conf import settings
from django.contrib.auth import authenticate, logout
from django.utils import timezone
from django.utils.translation import gettext
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.staticfiles import finders
from rest_framework.authtoken.models import Token
from rest_framework import viewsets
from rest_framework.views import Response as RestResponse
from rest_framework.decorators import action
from rest_framework import permissions as rest_permissions
from rest_framework import filters as rest_filters
from rest_framework.schemas import get_schema_view

import heliport.tasks
from .custom_views import HeliportLoginRequiredMixin, HeliportObjectView, HeliportProjectView, django_views
from .user_logic.user_information import authentication_backand_manager
from .user_logic.remote_login import download_ssh_key, delete_ssh_key
from .models import *
from .serializers import ProjectSerializer, UserSerializer, GroupSerializer
from .graph_generator import generate_graph, get_api_roots, get_search_urls, serialize_project_data
from .logger import logger
from .utils.handles import register_json_handle
from .permissions import HeliportProjectPermission, projects_of, StaffOrReadOnly
from utils.IDs import generate_project_id


class LoginView(django_views.TemplateView):
    template_name = 'heliport/user/heliport_login.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['redirect_to'] = self.request.GET.get('next', None)
        return context

    def post(self, request):
        username = request.POST.get('username')
        password = request.POST.get('password')
        user = authenticate(username=username, password=password)
        if user is None:
            logout(request)
            context = {
                'message': gettext('Login Failed'),
                'redirect_to': request.POST.get('redirect', None),
                **super().get_context_data()
            }
            return render(request, self.template_name, context)

        authentication_backand_manager.fetch_user(user)
        authentication_backand_manager.login(user, request)
        return redirect(request.POST.get('redirect', reverse('heliport:project_list')))


class LogoutView(django_views.View):
    def get(self, request):
        logout(request)
        return redirect('/')


class IndexView(django_views.View):
    def get(self, request):
        if request.user.is_authenticated:
            return redirect('heliport:project_list')
        return redirect('heliport:login')


class ProjectListView(HeliportLoginRequiredMixin, django_views.ListView):
    model = Project
    template_name = 'heliport/project/project_list.html'

    def get_queryset(self):
        user = self.request.user.heliportuser
        owned = user.project_set.filter(deleted__isnull=True)
        co_owned = user.co_owned_projects.filter(deleted__isnull=True)
        return owned.union(co_owned)


class ProjectGraphView(HeliportLoginRequiredMixin, UserPassesTestMixin, django_views.DetailView):
    template_name = 'heliport/project/project_graph.html'
    model = Project
    pk_url_kwarg = 'project'

    def test_func(self):
        user = self.request.user.heliportuser
        return user in self.get_object().members()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = context['project']
        self.request.session['activeHeliportProject'] = project.pk

        graph = generate_graph(self.object)
        context['graph'] = graph
        sections_count = len(list(graph))
        context['graph_size_large'] = 480 * sections_count
        context['graph_size_small'] = 370 * sections_count

        return context


class CreateProjectView(HeliportLoginRequiredMixin, django_views.View):
    def get(self, request):
        project = Project(owner=request.user.heliportuser)

        # assign group
        group_id = request.user.ldap_user.attrs['ou'][0]
        try:
            project.group = Group.objects.filter(group_id=group_id).get()
        except Group.DoesNotExist:
            new_group = Group(group_id=group_id)
            new_group.save()
            project.group = new_group
        project.save()

        project.hzdr_id = generate_project_id(request.user, project.group_id, project.created)
        project_graph_url = reverse('heliport:project_graph', kwargs={'project': project.pk})
        full_url = f'{request.scheme}://{request.get_host() }{project_graph_url}'
        if settings.REGISTER_HANDLES:
            project.handle = register_json_handle(project.hzdr_id, 'HELIPORT/Project', full_url).json().get('handle')
        project.save()

        return redirect(project_graph_url)


class UserSettingsView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = 'heliport/user/user_settings.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user
        context['tokens'] = Token.objects.filter(user=user)
        context['heliport_user'] = user.heliportuser

        return context

    def post(self, request, *args, **kwargs):
        user = request.user.heliportuser
        orcid = request.POST.get('orcid')
        orcid_link = 'https://orcid.org/'
        if orcid and not orcid.startswith(orcid_link):
            orcid = f'{orcid_link}{orcid}'

        user.orcid = orcid
        user.save()

        return redirect('heliport:user_settings')


class RemoteServerLoginView(HeliportLoginRequiredMixin, django_views.DetailView):
    model = LoginInfo
    template_name = 'heliport/user/remote_server_login.html'

    @staticmethod
    def validate_login(login, request):
        if request.user.heliportuser != login.user:
            return redirect(reverse('heliport:login'))

        logger.debug('check if all parents are connected')
        try:
            connect_first = login.get_unconnected_parents()
        except ValueError:
            return HttpResponse(f'<b> ERROR: </b> {gettext("infinite chain of ssh connections")}')
        if connect_first:
            request.session['message'] = f'{login.name} can not be connected. Please connect {connect_first[0]} first!'
            return redirect('heliport:remote_server_login', pk=connect_first[0].pk)

        return None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['message'] = self.request.session.get('message', '')
        self.request.session['message'] = ''
        return context

    def post(self, request, *args, **kwargs):
        logger.debug('user requested ssh key connection')
        login_info = self.get_object()
        error = self.validate_login(login_info, request)
        if error is not None:
            return error

        username = request.POST.get('username')
        password = request.POST.get('password')
        login_info.username = username
        login_info.save()

        download_ssh_key(login_info, username, password)
        return redirect(reverse('heliport:user_login_list'))


class RemoteServerLogoutView(HeliportLoginRequiredMixin, django_views.View):
    def get(self, request, *args, **kwargs):
        logger.debug('user requested ssh disconnect')
        login_info = get_object_or_404(LoginInfo, pk=kwargs.get('pk'))
        error = RemoteServerLoginView.validate_login(login_info, request)
        if error is not None:
            return error

        delete_ssh_key(login_info)
        return redirect(reverse('heliport:user_login_list'))


class LoginInfoView(HeliportLoginRequiredMixin, django_views.CreateView):
    model = LoginInfo
    template_name = 'heliport/user/login_infos.html'
    fields = ['name', 'type', 'username', 'password', 'key', 'machine', 'via', 'slurm']

    def get_success_url(self):
        return reverse('heliport:user_login_list')

    def get_context_data(self, **kwargs):
        user = self.request.user.heliportuser
        context = super().get_context_data(**kwargs)
        context['logins'] = user.logininfo_set.all()
        context['login_type'] = self.request.GET.get('login_type')
        context['login_types'] = LoginInfo.LoginTypes
        context['ssh_logins'] = user.logininfo_set.filter(type=LoginInfo.LoginTypes.SSH)
        context['update'] = False
        return context

    def post(self, request, *args, **kwargs):
        login_to_remove = request.POST.get('remove')
        if login_to_remove:
            login_info = get_object_or_404(LoginInfo, pk=login_to_remove)
            if login_info.user != self.request.user.heliportuser:
                return redirect(reverse('heliport:login'))
            login_info.delete()
            return redirect(self.get_success_url())
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        form.instance.user = self.request.user.heliportuser
        return super().form_valid(form)


class LoginInfoUpdateView(HeliportLoginRequiredMixin, django_views.UpdateView):
    model = LoginInfo
    template_name = 'heliport/user/login_infos.html'
    fields = ['name', 'type', 'username', 'password', 'key', 'machine', 'via', 'slurm']

    def get_success_url(self):
        return reverse('heliport:user_login_list')

    def get_context_data(self, **kwargs):
        user = self.request.user.heliportuser
        context = super().get_context_data(**kwargs)
        context['logins'] = user.logininfo_set.all()
        context['login_type'] = self.request.GET.get('login_type', self.object.type)
        context['login_types'] = LoginInfo.LoginTypes
        context['ssh_logins'] = user.logininfo_set.filter(type=LoginInfo.LoginTypes.SSH)
        context['update'] = True
        return context

    def form_valid(self, form):
        if form.instance.user != self.request.user.heliportuser:
            return redirect(reverse('heliport:login'))
        return super().form_valid(form)


class ProjectView(HeliportProjectView, django_views.DetailView):
    model = Project
    template_name = 'heliport/project/project_update.html'
    pk_url_kwarg = 'project'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['isOwner'] = self.object.owner == self.request.user.heliportuser
        context['groups'] = Group.objects.order_by('group_id')
        self.request.session['activeHeliportProject'] = self.object.pk
        context['update'] = False
        return context

    def post(self, request, *args, **kwargs):
        project_id = request.POST.get('remove')
        user_id = request.POST.get('add')
        delete_owner = self.request.POST.get('user')
        group = self.request.POST.get('group_to_import')

        project = self.get_object()

        if project_id:
            if project.owner == request.user.heliportuser:
                project.deleted = timezone.now()
                project.save()
            return redirect(reverse('heliport:project_list'))
        elif user_id:
            user = get_object_or_404(HeliportUser, pk=int(user_id))
            if project.owner != user:
                project.co_owner.add(user)

        elif delete_owner:
            project.co_owner.remove(int(delete_owner))

        elif group:
            for db_user in authentication_backand_manager.get_users_by_group(group):
                if db_user != project.owner:
                    project.co_owner.add(db_user)

        return redirect('heliport:project', project=project.pk)


class ProjectUpdateView(HeliportProjectView, django_views.UpdateView):
    model = Project
    template_name = 'heliport/project/project_update.html'
    fields = ['project_name', 'description', 'group']
    pk_url_kwarg = 'project'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['isOwner'] = self.object.owner == self.request.user.heliportuser
        context['groups'] = Group.objects.order_by('group_id')
        self.request.session['activeHeliportProject'] = self.object.pk
        context['update'] = True
        return context

    def get_success_url(self):
        return reverse('heliport:project', kwargs={'project': self.object.pk})


def get_users_autocomplete(request, **kwargs):
    query = request.GET.get('q')

    if query:
        sql_query = HeliportUser.objects.filter(display_name__icontains=query)
    else:
        sql_query = HeliportUser.objects.all()

    results = [{'name': guy.display_name, 'pk': guy.pk} for guy in sql_query]
    return JsonResponse(results, safe=False)


def get_project_list(request):
    if request.user.is_authenticated:
        user = request.user.heliportuser
        p = list(user.project_set.all().union(user.co_owned_projects.all()))
        data = [{'name': obj.project_name, 'pk': obj.pk, 'hzdr_id': obj.hzdr_id} for obj in p if obj.deleted is None]
        return JsonResponse(data, safe=False)
    return JsonResponse([], safe=False)


class ProjectSerializationView(HeliportProjectView, django_views.DetailView):
    model = Project
    pk_url_kwarg = 'project'

    def get(self, request, **kwargs):
        return JsonResponse(serialize_project_data(self.get_object()), json_dumps_params={'indent': 2})


class FaviconView(django_views.View):
    def get(self, request, **kwargs):
        icon = finders.find('img/favicon.ico')
        with open(icon, 'rb') as f:
            response = HttpResponse(f.read(), content_type='image/ico')
        return response


class SearchBaseView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = 'heliport/search/search_base.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['search_urls'] = list(get_search_urls())
        context['query'] = self.request.GET.get('q', '')
        return context


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = 'heliport/search/search.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context['query'] = query

        project_results = set()
        group_results = set()
        user_results = set()
        login_info_results = set()

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            project_results.update(Project.objects.filter(
                (Q(owner=user) | Q(co_owner=user)) & Q(deleted__isnull=True) &
                (Q(project_id=num) |
                 Q(hzdr_id__icontains=word) |
                 Q(project_name__icontains=word) |
                 Q(owner__display_name__icontains=word) |
                 Q(co_owner__display_name__icontains=word) |
                 Q(description__icontains=word) |
                 Q(group=word))
            ))

            group_results.update(Group.objects.filter(group_id__icontains=word))

            user_results.update(HeliportUser.objects.filter(
                Q(display_name__icontains=word) |
                Q(user_id=num)
            ))

            login_info_results.update(LoginInfo.objects.filter(
                Q(user=user) &
                (
                    Q(username__icontains=word) |
                    Q(machine__icontains=word) |
                    Q(name__icontains=word) |
                    Q(type__icontains=word)
                )
            ))

        context['project_results'] = project_results
        context['group_results'] = group_results
        context['user_results'] = user_results
        context['login_info_results'] = login_info_results
        return context


class ManualUserDownloadView(HeliportLoginRequiredMixin, django_views.View):
    def get(self, request, **kwargs):
        result = heliport.tasks.import_user_data()
        return HttpResponse(f'OK. {result} users imported')


##########################################
#               REST API                 #
##########################################

class ProjectViewSet(viewsets.ModelViewSet):
    """
    Showing the most general HELIPORT project properties.
    Request more detailed information by appending urls/ to the url
    Go to a specific project by appending its id/ to the url
    For more information on how to authenticate look in HELIPORT (user > settings)
    """
    serializer_class = ProjectSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportProjectPermission]
    filter_backends = [rest_filters.SearchFilter]
    search_fields = ['hzdr_id', 'project_name', 'description', 'owner__display_name']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return projects_of(user)

    @action(detail=False)
    def urls(self, request):
        api_roots = [f'{ request.scheme }://{ request.get_host() }{url}' for url in get_api_roots()]
        return RestResponse(api_roots)


class UserViewSet(viewsets.ModelViewSet):
    """Showing HELIPORT user model."""
    permission_classes = [rest_permissions.IsAuthenticated, StaffOrReadOnly]
    filterset_fields = ['user_id', 'authentication_backend_id', 'display_name']

    queryset = HeliportUser.objects.all()
    serializer_class = UserSerializer


class GroupViewSet(viewsets.ModelViewSet):
    """Showing user groups."""
    permission_classes = [rest_permissions.IsAuthenticated, StaffOrReadOnly]

    queryset = Group.objects.all()
    serializer_class = GroupSerializer


# the schema view read by redoc
SchemaVew = get_schema_view(
    title='HELIPORT',
    description='API for all things …',
    version='0.0.3'
)


class RedocView(django_views.TemplateView):
    template_name = 'heliport/api/redoc.html'
