from pathlib import Path
import logging

import environ
from django.utils.translation import ugettext_lazy as translate
from django_auth_ldap.config import LDAPSearch, GroupOfNamesType
from celery.schedules import crontab
import ldap

# ==== setup setting creation ====
BASE_DIR = Path(__file__).resolve().parent.parent
env = environ.Env()
env.read_env()

# ==== meta settings (add shorthand for important settings here) ====
# list of custom apps in heliport-system
# add app urls also in include_urls.py
_HELIPORT_SYSTEM_APPS = [
    'heliport',
    'gate_connection',
    'version_control',
    'data_management_plan',
    'documentation',
    'data_source',
    'archive',
    'publication',
    'cwl_execution',
    'hzdr_about',
    'digital_objects',
]

# configure for local development
_LOCAL_EXECUTION = 'LOCAL_EXECUTION' in env and env('LOCAL_EXECUTION') == 'True'
# remove security constraints to allow connection with other devices to a local development server
# add your ip to allowed hosts, set below to "True and _LOCAL_EXECUTION" and run manage.py runserver <your ip>
_TEST_LOCAL_NETWORK = False and _LOCAL_EXECUTION

# ==== other manually added settings ====
# don't use cookies to remember users -> login every time
SESSION_EXPIRE_AT_BROWSER_CLOSE = True
# generate project identifiers with this prefix
HZDR_ID_PREFIX = 'HZDR'
# name of SSH key saved on connected computers
SSH_KEY_NAME = 'heliport_rsa' if not _LOCAL_EXECUTION else 'heliport_rsa_local_execution'
# use ssh connection for executing commands that need to be run in local network
DATA_SYNC_VIA_SSH = 'uts.hzdr.de' if _LOCAL_EXECUTION else None
# command to get all users of a group separated by \n\n, users that don't have "cn: " and "displayName: " are ignored
_get_ldap_users = 'ldapsearch -D "" -h ldap.fz-rossendorf.de -b "ou=users,ou=db,ou=it,o=fsr,dc=de"'
GET_USERS_BY_GROUP = f'{_get_ldap_users} "ou={{group}}"'
# command to get all groups, one group per line
_get_ldap_groups = 'ldapsearch -D "" -h ldap.fz-rossendorf.de -b "ou=groups,ou=db,ou=it,o=fsr,dc=de" "businessCategory=1003"'
GET_GROUPS_COMMAND = fr'{_get_ldap_groups} | grep "^description: " | sed "s/^description: \([^ ]\+\) .*$/\1/g"'

# ==== HELIPORT graph configuration ====
GRAPH = {
    'Project': ['project_configuration', 'gate_connection'],
    'Systems': ['version_control', 'data_management_plan', 'documentation', 'digital_objects'],
    'Resources': ['data_source', 'arbitrary_cwl', 'tool', 'workflow'],
    'Automation': ['job'],
    'Results': ['archive', 'publication'],
}

DEPENDENCIES = {
    'version_control': ['project_configuration'],
    'data_management_plan': ['project_configuration'],
    'documentation': ['project_configuration'],
    'data_source': ['project_configuration'],
    'archive': ['project_configuration'],
    'publication': ['project_configuration'],
    'job': ['project_configuration'],
    'tool': ['project_configuration'],
    'workflow': ['project_configuration'],
    'arbitrary_cwl': ['project_configuration'],
    'digital_objects': ['project_configuration'],
}

# job execution settings
USE_SRUN = True
SRUN_TIMEOUT = 10
USE_CELERY = not _LOCAL_EXECUTION
SSH_DATA_TRANSFER_BATCH_SIZE = 10_000  # in bytes
DEFAULT_JOB_SETUP_COMMAND = '''ml python'''

# handle settings
REGISTER_HANDLES = not _LOCAL_EXECUTION
HANDLE_PREFIX = '20.500.12865'
handle_address = 'handle.hzdr.de'
HANDLE_API_URL = f'https://{handle_address}/api/handles'
HANDLE_CERTIFICATE_PATH = '/etc/ssl/certs/heliport_handle_admin.crt'
HANDLE_PRIVATE_KEY_PATH = '/etc/ssl/private/heliport_handle_admin.key'
HANDLE_DESCRIPTION = ('This handle is generated by HELIPORT at Helmholtz-Zentrum Dresden-Rossendorf. '
                      'HELIPORT is a system in its early prototyping phase and not yet available to the public. '
                      'If you are at HZDR and you have access rights, you can follow the URL below. ')

# ==== derived and django generated settings ====
# Application definition
INSTALLED_APPS = [
                     'django.contrib.admin',
                     'django.contrib.auth',
                     'django.contrib.contenttypes',
                     'django.contrib.sessions',
                     'django.contrib.messages',
                     'django.contrib.staticfiles',
                     'rest_framework',  # for REST API
                     'rest_framework.authtoken',  # for using Token authentication
                     'django_filters',  # for adding query arguments to rest url
                 ] + _HELIPORT_SYSTEM_APPS

MIDDLEWARE = [
    # auto generated middleware:
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    # custom middleware:
    'django.middleware.locale.LocaleMiddleware',  # for translation
    'heliport.middleware.never_cache_middleware.DisableClientSideCachingMiddleware'  # reload page every time
]

ROOT_URLCONF = 'heliport_project.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [BASE_DIR / "core_resources/templates"],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'heliport_project.wsgi.application'

if _LOCAL_EXECUTION:
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': BASE_DIR / 'db.sqlite3',
        }
    }
else:
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql',
            'NAME': 'guidance',
            'USER': env('DB_USER'),
            'PASSWORD': env('DB_PASSWORD'),
            'HOST': 'vlspgdb0.fz-rossendorf.de',
            'PORT': '5432',
        }
    }

# Password validation
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

# Internationalization
LANGUAGES = [
    ('de', translate('German')),
    ('en', translate('English')),
]

LOCALE_PATHS = [BASE_DIR / 'locale']
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True

# Static files (CSS, JavaScript, Images)
STATIC_URL = '/static/'
STATIC_ROOT = 'static'
STATICFILES_DIRS = [
    BASE_DIR / 'core_resources/static'
]

# ldap
AUTHENTICATION_BACKENDS = [
    'django_auth_ldap.backend.LDAPBackend',
    'django.contrib.auth.backends.ModelBackend'
]

PROFILE_UPDATE = False
PROFILE_DELETE = False

if not _LOCAL_EXECUTION:
    AUTH_LDAP_SERVER_URI = "ldap://ldap.fz-rossendorf.de"
else:
    AUTH_LDAP_SERVER_URI = "ldap://localhost:10389"
AUTH_LDAP_BIND_DN = ""
AUTH_LDAP_BIND_PASSWORD = ""
AUTH_LDAP_USER_SEARCH = LDAPSearch("ou=users,ou=db,ou=it,o=fsr,dc=de", ldap.SCOPE_SUBTREE, "(uid=%(user)s)")
AUTH_LDAP_GROUP_SEARCH = LDAPSearch(
    "ou=groups,ou=db,ou=it,o=fsr,dc=de", ldap.SCOPE_SUBTREE, "(objectClass=top)"
)
AUTH_LDAP_GROUP_TYPE = GroupOfNamesType()
# AUTH_LDAP_MIRROR_GROUPS = True  # download groups of user that loggs in

AUTH_LDAP_USER_ATTR_MAP = {
    "first_name": "givenName",
    "last_name": "sn",
    'email': 'mail'
}

# logging
if not _LOCAL_EXECUTION:
    LOG_FILENAME = '/var/log/heliport/heliport_base.log'
else:
    LOG_FILENAME = 'heliport_base.log'

LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s :-: %(message)s'
LOG_FORMAT_TIME = '%m/%d/%Y %I:%M:%S %p'
LOG_LEVEL = logging.DEBUG if _LOCAL_EXECUTION else logging.INFO
LOG_TO_CONSOLE = _LOCAL_EXECUTION

# rest framework
REST_FRAMEWORK = {
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'PAGE_SIZE': 10,
    'DEFAULT_AUTHENTICATION_CLASSES': [
        'rest_framework.authentication.TokenAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    ],
    'DEFAULT_PERMISSION_CLASSES': [
        'rest_framework.permissions.IsAuthenticated',
    ],
    'DEFAULT_FILTER_BACKENDS': [
        'django_filters.rest_framework.DjangoFilterBackend',
    ]
}

# Celery Configuration Options
BROKER_URL = 'pyamqp://localhost/vlsdms' if _LOCAL_EXECUTION else env('CELERY_BROKER')
CELERY_RESULT_BACKEND = 'amqp'
CELERY_TASK_RESULT_EXPIRES = 18000  # 5 hours.

CELERY_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TASK_SERIALIZER = 'json'
CELERY_ACCEPT_CONTENT = ['json']
CELERY_TIMEZONE = 'Europe/Berlin'

CELERYBEAT_SCHEDULE = {
    'healthcheck': {
        'task': 'heliport_project.celery.add',
        'schedule': crontab(hour='*/1'),
        'args': (5, 12),
    },
    'daily_gate_sync': {
        'task': 'gate_connection.tasks.update_gate_projects',
        'schedule': crontab(minute=30, hour=2),
    },
    'daily_ldap_sync': {
        'task': 'heliport.tasks.import_user_data',
        'schedule': crontab(minute=10, hour=2),
    },
    'periodic_job_execution': {
        'task': 'cwl_execution.tasks.run_periodic_jobs',
        'schedule': crontab(minute=50, hour=2),
    }
}

# ==== SECURITY_CRITICAL_SETTINGS ====
SECRET_KEY = env("SECRET_KEY")
DEBUG = _LOCAL_EXECUTION
if not _LOCAL_EXECUTION:
    ALLOWED_HOSTS = ["vlsdms", "vlsdms.fz-rossendorf.de", "127.0.0.1"]
else:
    ALLOWED_HOSTS = []
CSRF_COOKIE_SECURE = not _TEST_LOCAL_NETWORK
SESSION_COOKIE_SECURE = not _TEST_LOCAL_NETWORK
SECURE_SSL_REDIRECT = not _LOCAL_EXECUTION
