from django.db.models import Q
from django.shortcuts import reverse, get_object_or_404
from rest_framework import viewsets
from rest_framework import permissions as rest_permissions

from heliport.permissions import HeliportObjectPermission, is_object_member
from heliport.custom_views import HeliportProjectView, HeliportLoginRequiredMixin, django_views
from .models import *
from .serializers import PublicationSerializer


class PublicationView(HeliportProjectView, django_views.CreateView):
    model = Publication
    template_name = "publication/publications.html"
    fields = ["description"]

    def form_valid(self, form):
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        form.instance.project = project
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("publication:list", kwargs=self.kwargs)


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = "publication/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context["query"] = query

        publication_results = set()

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            publication_results.update(Publication.objects.filter(
                Q(publication_id=num) |
                Q(description__icontains=word)
            ))

        context["publication_results"] = publication_results
        return context


##########################################
#               REST API                 #
##########################################

class PublicationViewSet(viewsets.ModelViewSet):
    """Publication"""
    serializer_class = PublicationSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['publication_id', 'project']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return Publication.objects.filter(is_object_member(user))
