import abc
import typing
import io
import subprocess
import paramiko
from cryptography.fernet import Fernet
import json


class CommandResults:
    def __init__(self, stdout: str, stderr: str, status: int):
        self.stdout = stdout
        self.stderr = stderr
        self.status = status

    @property
    def stdout_stderr(self) -> typing.Tuple[str, str]:
        return self.stdout, self.stderr

    @property
    def status_result(self) -> typing.Tuple[int, str]:
        return self.status, self.result

    @property
    def result(self) -> str:
        if self.stdout and self.stderr:
            return self.stdout + self.stderr
        elif self.stdout:
            return self.stdout
        elif self.stderr:
            return self.stderr
        else:
            return ''

    @property
    def last_line(self) -> str:
        return self.result.splitlines()[-1]

    @property
    def last_complete_line(self) -> str:
        for line in reversed(self.result.splitlines(keepends=True)):
            clean_line = line.rstrip('\r\n')
            if line != clean_line:
                return clean_line
        return ''

    def __str__(self):
        return self.result

    def extract_http_content(self):
        http = self.stdout.split('\r\n\r\n', 1)
        if len(http) < 2:
            raise ValueError(f'command does not produce http response\n{self.stdout}\n{self.stderr}')
        header, data = http
        return data

    def extract_http_json(self):
        return json.loads(self.extract_http_content())


class CommandExecutor:
    def __init__(self,
                 allow_shell_syntax: bool = True,
                 shell: bool = False,
                 local: bool = True,
                 host: str = None,
                 username: str = None,
                 password: str = None,
                 key: str = None,
                 import_local_keys: bool = False,
                 key_password_callback: typing.Callable[[], str] = None,
                 indirection: 'CommandExecutor' = None,
                 auto_connect: bool = True):
        self.allow_shell_syntax = allow_shell_syntax
        self.shell = shell
        self.local = local
        self.host = host
        self.username = username
        self.password = password
        self.key = key
        self.import_local_keys = import_local_keys
        self.key_password_callback = key_password_callback
        self.indirection = indirection

        self.to_close = []

        self.connected = False
        if auto_connect:
            self.connect()

    def __enter__(self):
        if not self.connected:
            self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        while self.to_close:
            self.to_close.pop().close()

        self.connected = False

    @abc.abstractmethod
    def execute(self, command: str) -> CommandResults:
        pass

    @abc.abstractmethod
    def connect(self):
        pass

    @abc.abstractmethod
    def file(self, path: str, mode: str = 'r'):
        pass


class LocalCommandExecutor(CommandExecutor):
    def __init__(self,
                 allow_shell_syntax=True,
                 auto_connect=True):
        super().__init__(allow_shell_syntax=allow_shell_syntax,
                         shell=False,
                         local=True,
                         auto_connect=auto_connect)

    def connect(self):
        self.connected = True

    def file(self, path, mode='r'):
        f = open(path, mode)
        self.to_close.append(f)
        return f

    def execute(self, command) -> CommandResults:
        result = subprocess.run(command, shell=self.allow_shell_syntax, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return CommandResults(result.stdout.decode('utf-8'), result.stderr.decode('utf-8'), result.returncode)


class SSHCommandExecutor(CommandExecutor):
    @staticmethod
    def encrypt(value: str, key: str):
        encryption = Fernet(key.encode())
        encrypted_value = encryption.encrypt(value.encode())
        return encrypted_value.decode()

    def __init__(self,
                 host: str = None,
                 username: str = None,
                 password: str = None,
                 key: str = None,
                 import_local_keys: bool = False,
                 key_password_callback: typing.Callable[[], str] = None,
                 indirection: 'CommandExecutor' = None,
                 auto_connect: bool = True,
                 missing_host_key_policy=None,
                 set_shell_to_true: bool = False):
        self.ssh = None
        self.file_server = None
        if missing_host_key_policy is None:
            missing_host_key_policy = paramiko.AutoAddPolicy()
        self.missing_host_key_policy = missing_host_key_policy

        super().__init__(allow_shell_syntax=True,
                         shell=set_shell_to_true,
                         local=False,
                         host=host,
                         username=username,
                         password=password,
                         key=key,
                         import_local_keys=import_local_keys,
                         key_password_callback=key_password_callback,
                         indirection=indirection,
                         auto_connect=auto_connect)

    def execute(self, command: str, command_input: str = None) -> CommandResults:
        if not self.connected:
            raise ValueError('please call connect() before executing commands')

        stdin, stdout, stderr = self.ssh.exec_command(command)
        if command_input is not None:
            print(command_input, file=stdin)
        return CommandResults(stdout.read().decode(),
                              stderr.read().decode(),
                              stdout.channel.recv_exit_status())

    @property
    def decrypted_key(self) -> typing.Optional[paramiko.RSAKey]:
        if self.key is None:
            return None

        if self.key_password_callback is not None:
            f = Fernet(self.key_password_callback())
            decrypted_private_key = f.decrypt(self.key.encode())
        else:
            decrypted_private_key = self.key

        private_key_string = io.StringIO(decrypted_private_key.decode())
        private_key = paramiko.RSAKey.from_private_key(private_key_string)
        return private_key

    def connect(self, max_depth: int = 10):
        if max_depth <= 0:
            raise ValueError('max_depth is reached - this may be caused by cyclical ssh connections')

        self.setup_new_ssh()
        self.ssh.connect(self.host,
                         username=self.username,
                         password=self.password,
                         pkey=self.decrypted_key,
                         sock=self.get_parent_channel(max_depth),
                         look_for_keys=self.import_local_keys)

        self.to_close.append(self.ssh)
        self.connected = True

    def get_parent_channel(self, max_depth: int):
        if isinstance(self.indirection, SSHCommandExecutor):
            if not self.indirection.connected:
                self.indirection.connect(max_depth=max_depth - 1)
                self.to_close.append(self.indirection)

            channel = self.indirection.ssh.get_transport().open_session()
            channel.exec_command(f'nc {self.host} {paramiko.config.SSH_PORT}')
            return channel
        return None

    def setup_new_ssh(self):
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(self.missing_host_key_policy)
        if self.import_local_keys:
            self.ssh.load_system_host_keys()

    def file(self, path: str, mode: str = 'r'):
        if not self.connected:
            raise ValueError('please call connect() before opening files')

        if self.file_server is None:
            self.file_server = self.ssh.open_sftp()
        f = self.file_server.file(path, mode)
        self.to_close.append(f)
        return f


class ShellSSHCommandExecutor(SSHCommandExecutor):
    def __init__(self,
                 host: str = None,
                 username: str = None,
                 password: str = None,
                 key: str = None,
                 import_local_keys: bool = False,
                 key_password_callback: typing.Callable[[], str] = None,
                 indirection: 'CommandExecutor' = None,
                 auto_connect: bool = True,
                 default_result_callback: typing.Callable[[str], None] = None,
                 missing_host_key_policy=None):
        self.stdin = None
        self.stdout = None
        self.command_separator = 'end of stdOUT buffer. finished with exit status:'
        self.default_result_callback = default_result_callback
        super().__init__(host=host,
                         username=username,
                         password=password,
                         key=key,
                         import_local_keys=import_local_keys,
                         key_password_callback=key_password_callback,
                         indirection=indirection,
                         auto_connect=auto_connect,
                         missing_host_key_policy=missing_host_key_policy,
                         set_shell_to_true=True)

    def connect(self, max_depth: int = 10):
        super().connect(max_depth)
        channel = self.ssh.invoke_shell(width=1000)
        channel.settimeout(1)
        self.stdin = channel.makefile('wb')
        self.stdout = channel.makefile('r')

    def read_output(self) -> str:
        result = b''
        while True:
            try:
                result += self.stdout.read(1)
            except paramiko.channel.socket.timeout:
                break
        try:
            return result.decode()
        except UnicodeDecodeError:
            return '[UnicodeDecodeError shell probably paused half way through sending a multibyte UTF-8 character]'

    @property
    def echo_command_separator(self):
        return f'echo {self.command_separator} $?'

    def execute(self, command: str, output_callback: typing.Callable[[str], None] = None):
        if not self.connected:
            raise ValueError('please call connect() before executing commands')
        if output_callback is None:
            output_callback = self.default_result_callback

        to_write = f'{command.strip()}; {self.echo_command_separator}\n'
        self.stdin.write(to_write)
        self.stdin.flush()

        exit_code, result = self.await_command_result(output_callback)
        return CommandResults(result, '', exit_code)

    def await_command_result(self, output_callback=None):
        total_result = ''
        exit_code = -1
        waiting = True
        while waiting:
            result = self.read_output()
            batch_result = ''
            for line in result.splitlines(keepends=True):
                if line.startswith(self.command_separator):
                    exit_code = int(line.rsplit(maxsplit=1)[1])
                    waiting = False
                elif line.rstrip('\n\r').endswith(self.echo_command_separator):
                    line_separator = line[len(line.rstrip('\n\r')):]
                    batch_result += line.rsplit(';', maxsplit=1)[0] + line_separator
                else:
                    batch_result += line

            total_result += batch_result
            if output_callback is not None:
                output_callback(batch_result)

        return exit_code, total_result


def create_command_executor(
        allow_shell_syntax: bool = True,
        shell: bool = False,
        host: str = None,
        username: str = None,
        password: str = None,
        key: str = None,
        import_local_keys: bool = False,
        key_password_callback: typing.Callable[[], str] = None,
        indirection: CommandExecutor = None,
        auto_connect: bool = True,
        default_result_callback: typing.Callable[[str], None] = None,
        missing_host_key_policy=None) -> CommandExecutor:
    if host is not None:
        assert username is not None, 'please specify the username'

    if host is None:
        error_message = 'set host or change value of'
        executor = LocalCommandExecutor(
            allow_shell_syntax=allow_shell_syntax,
            auto_connect=auto_connect
        )
    elif shell:
        error_message = 'shell does not support'
        executor = ShellSSHCommandExecutor(
            host=host,
            username=username,
            password=password,
            key=key,
            import_local_keys=import_local_keys,
            key_password_callback=key_password_callback,
            indirection=indirection,
            auto_connect=auto_connect,
            missing_host_key_policy=missing_host_key_policy,
            default_result_callback=default_result_callback
        )
    else:
        error_message = 'ssh executor with shell=False does not support'
        executor = SSHCommandExecutor(
            host=host,
            username=username,
            password=password,
            key=key,
            import_local_keys=import_local_keys,
            key_password_callback=key_password_callback,
            indirection=indirection,
            auto_connect=auto_connect,
            missing_host_key_policy=missing_host_key_policy
        )
    parameters = {
        'allow_shell_syntax': allow_shell_syntax,
        'shell': shell,
        'host': host,
        'username': username,
        'password': password,
        'key': key,
        'import_local_keys': import_local_keys,
        'key_password_callback': key_password_callback,
        'indirection': indirection,
        'connected': auto_connect
    }
    for name, value in parameters.items():
        if getattr(executor, name) != value:
            raise ValueError(f'Invalid configuration: {error_message} {name}={value}')
    return executor
