import json
from django.test import TestCase
from django.contrib.auth.models import User
from django.shortcuts import reverse

from heliport.models import Project, HeliportUser
from .models import VersionControl


class PublicationTests(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username='testuser')
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.client.force_login(self.auth_user)
        self.project = Project.objects.create(owner=self.user)

    def test_list(self):
        vc, is_new = VersionControl.objects.get_or_create(project=self.project)

        response = self.client.get(reverse('version_control:list', kwargs={'project': self.project.pk}))
        self.assertEquals(200, response.status_code)
        self.assertEquals(response.context_data['project'].versioncontrol_set.get(pk=vc.pk), vc)

        response = self.client.get(reverse('version_control:update', kwargs={'project': self.project.pk, 'pk': vc.pk}))
        self.assertEquals(200, response.status_code)
        self.assertEquals(response.context_data['project'].versioncontrol_set.get(pk=vc.pk), vc)

    def test_create(self):
        self.client.post(reverse('version_control:list', kwargs={'project': self.project.pk}),
                         data={
                             'name': 'test_version_control321',
                             'link': 'https://www.google.com'
                         })
        test_obj = self.project.versioncontrol_set.filter(name='test_version_control321').first()
        self.assertIsNotNone(test_obj)
        self.assertEquals('https://www.google.com', test_obj.link)

    def test_update(self):
        vc, is_new = VersionControl.objects.get_or_create(project=self.project,
                                                          name='test_version_control456',
                                                          link='http://www.google.com')
        response = self.client.post(reverse('version_control:update', kwargs={'project': self.project.pk, 'pk': vc.pk}),
                                    data={
                                        'name': 'test_version_control789',
                                        'link': 'https://www.bing.com'
                                    }, follow=True)
        self.assertEquals(200, response.status_code)
        vc = response.context_data['project'].versioncontrol_set.get(pk=vc.pk)
        self.assertEquals('test_version_control789', vc.name)
        self.assertEquals('https://www.bing.com', vc.link)

    def test_delete(self):
        vc, is_new = VersionControl.objects.get_or_create(project=self.project)
        response = self.client.post(reverse('version_control:list', kwargs={'project': self.project.pk}),
                                    data={
                                        'remove': vc.pk
                                    }, follow=True)
        self.assertEquals(200, response.status_code)
        self.assertIsNone(response.context_data['project'].versioncontrol_set.filter(pk=vc.pk).first())

    def test_add_protocol_create(self):
        self.client.post(reverse('version_control:list', kwargs={'project': self.project.pk}),
                         data={
                             'name': 'test_version_controlA',
                             'link': 'www.google.com'
                         })
        test_obj = self.project.versioncontrol_set.filter(name='test_version_controlA').first()
        self.assertIsNotNone(test_obj)
        self.assertEquals('https://www.google.com', test_obj.link)

    def test_add_protocol_update(self):
        vc, is_new = VersionControl.objects.get_or_create(project=self.project,
                                                          name='test_version_controlB',
                                                          link='http://www.google.com')
        response = self.client.post(reverse('version_control:update', kwargs={'project': self.project.pk, 'pk': vc.pk}),
                                    data={
                                        'name': 'test_version_controlB',
                                        'link': 'www.bing.com'
                                    }, follow=True)
        self.assertEquals(200, response.status_code)
        vc = response.context_data['project'].versioncontrol_set.get(pk=vc.pk)
        self.assertEquals('https://www.bing.com', vc.link)


class SearchAndAPITest(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username='testuser')
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.project, is_new = Project.objects.get_or_create(owner=self.user)
        self.client.force_login(self.auth_user)

    def test_access_search(self):
        response = self.client.get(reverse('version_control:search'))
        self.assertEquals(200, response.status_code)

    def test_version_control_findable(self):
        VersionControl.objects.create(project=self.project, name='VERSION_CONTROL321')
        response = self.client.get(f'{reverse("version_control:search")}?q=321')
        self.assertContains(response, 'VERSION_CONTROL321')

    def test_api(self):
        VersionControl.objects.create(project=self.project, name='VERSION_CONTROL654')
        response = self.client.get(reverse('version_control:api_root'))
        self.assertEquals(200, response.status_code)
        urls = json.loads(response.content.decode())
        response = self.client.get(urls['version_control'])
        self.assertEquals(200, response.status_code)
        self.assertContains(response, 'VERSION_CONTROL654')
