from django.db.models import Q
from django.shortcuts import redirect, get_object_or_404, reverse
from rest_framework import viewsets
from rest_framework import permissions as rest_permissions

from .serializers import VersionControlSerializer
from heliport.permissions import HeliportObjectPermission, is_object_member
from heliport.custom_views import HeliportObjectView, HeliportProjectView, HeliportLoginRequiredMixin, django_views
from .models import *


class VersionControlView(HeliportProjectView, django_views.CreateView):
    model = VersionControl
    template_name = "version_control/version_control.html"
    fields = ["name", "link"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        version_control_id = request.POST.get("remove")
        if version_control_id:
            version_control = get_object_or_404(VersionControl, pk=version_control_id)
            version_control.delete()
            return redirect(reverse("version_control:list", kwargs=self.kwargs))
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        form.instance.project = project
        if not form.instance.link.startswith("http://") and not form.instance.link.startswith("https://"):
            form.instance.link = "https://" + form.instance.link
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("version_control:list", kwargs=self.kwargs)


class VersionControlUpdateView(HeliportObjectView, django_views.UpdateView):
    model = VersionControl
    template_name = "version_control/version_control.html"
    fields = ["name", "link"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["update"] = True
        return context

    def form_valid(self, form):
        if not form.instance.link.startswith("http://") and not form.instance.link.startswith("https://"):
            form.instance.link = "https://" + form.instance.link
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("version_control:list", kwargs={'project': self.object.project.pk})


class SearchView(HeliportLoginRequiredMixin, django_views.TemplateView):
    template_name = "version_control/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get('q', '')
        context["query"] = query

        version_control_results = set()

        for word in query.split():
            num = int(word) if word.isnumeric() else -1

            version_control_results.update(VersionControl.objects.filter(
                (Q(project__owner=user) | Q(project__co_owner=user)) & Q(project__deleted__isnull=True) &
                (
                    Q(version_control_id=num) |
                    Q(name__icontains=word) |
                    Q(link__icontains=word)
                )
            ))

        context["version_control_results"] = version_control_results
        return context


##########################################
#               REST API                 #
##########################################

class VersionControlViewSet(viewsets.ModelViewSet):
    """Version Control"""
    serializer_class = VersionControlSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]
    filterset_fields = ['version_control_id', 'name', 'project']

    def get_queryset(self):
        user = self.request.user.heliportuser
        return VersionControl.objects.filter(is_object_member(user))
